from __future__ import annotations

from abc import abstractmethod
from datetime import datetime, timedelta

import numpy as np
import scipy.optimize

import skretrieval.time.eci

from ..platform.platformlocator import PlatformLocation

# ----------------------------------------------------------------------------
#          class SatelliteBase                                  2004-11-23*/
#   SatelliteBase is a base class for artificial satellites orbiting the Earth.
#   The class depends upon derived classes to provide the calculation of eciposition and ecivelocity
#   at any instant in time. The class uses Earth Centred Intertial coordinates using the true equator and true
#   equinox.  Individual satellite predictors should ensure they are consistent
#   with true equator and mean equinox in all calculations.
# --------------------------------------------------------------------------*/


class SatelliteBase(PlatformLocation):
    """
    The SatelliteBase is a base class for artificial satellites orbiting the Earth. The class derives from
    :class:`~.PlatformLocation` which allows the SatelliteBase class and its derivatives to be  used by the *skretrieval*
    package to calculate the location of instruments mounted on satellites. The ECI-TEME coordinate system is used for
    storage of satellite position and velocity. There is some looseness in the definition of Earth Centered Inertial.
    We use the definition of ECI employed by the SGP4 orbital propagator code which is a Celestial Reference system using
    the true equator and mean equinox, TEME,  i.e. precession is accounted for but nutation is not. This system is not the same as
    the GCRS system commonly used in astronomy applications today so care must be taken when transforming coordinate systems. The
    conversion to and from ECI-TEME is provided by the `eci` functions in package `sktime` which have been checked to centimeter
    precision.

    The SatelliteBase class overrides and implements the following three functions of parent class :class:`~.PlatformLocation`

    #.  :attr:`~PlatformLocation.position`
    #.  :attr:`~PlatformLocation.velocity`
    #.  :meth:`~PlatformLocation.update_position`

    Note that the above three functions return answers in the geographic geocentric system and not the eci system.

    Child classes that implement specific satellite orbit propagator methods and derive from *SatelliteBase*  must
    implement the following abstract methods:

    #.  :meth:`~SatelliteBase.update_eci_position`
    #.  :attr:`~SatelliteBase.period`

    """

    def __init__(self):
        super().__init__()
        self._m_time: datetime = (
            None  # The time used in the last call to self.update_eci_position
        )
        self._temptime: datetime = (
            None  # Local variable used for internal calculations, eg _zcomponent
        )
        self._m_orbit_number_at_start_time: int = (
            None  # Orbit number at time *_m_start_time_of_orbit*
        )
        self._m_start_time_of_orbit: datetime = None  # The Start time of orbitnumber.
        self._m_ecivelocity: np.ndarray = np.zeros(
            [3]
        )  # The ECI velocity in m/s generated in the last call to self.update_eci_position
        self._m_ecilocation: np.ndarray = np.zeros(
            [3]
        )  # The ECI position in meters generated in the last call to self.update_eci_position

    # -----------------------------------------------------------------------------
    #           time
    # -----------------------------------------------------------------------------
    @property
    def time(self) -> datetime:
        """
        Returns the UTC time of the current state vector

        Returns
        -------
        datetime.datetime
            The time of the current state vector
        """
        return self._m_time

    # -----------------------------------------------------------------------------
    #           geoposition
    # -----------------------------------------------------------------------------
    @property
    def position(self) -> np.ndarray:
        """
        Returns the position of the satellite in meters in the geocentric geographic ECEF
        coordinate system.

        Returns
        -------
        np.ndarray[3]
            The three element array specifying (X,Y,Z) in meters
        """
        return skretrieval.time.eci.eciteme_to_itrf(self._m_ecilocation, self._m_time)

    # -----------------------------------------------------------------------------
    #           geovelocity
    # -----------------------------------------------------------------------------
    @property
    def velocity(self) -> np.ndarray:
        """ECEFITRF/GEO coordinate system.

        ..  note::

            The velocity is the same vector as calculated for the ECI/TEME frame. It does not include the component
            due to the rotation of the Earth.

        Returns
        -------
        np.ndarray[3]
            The three element array specifying (X,Y,Z) in meters
        """
        return skretrieval.time.eci.eciteme_to_itrf(self._m_ecivelocity, self._m_time)

    # -----------------------------------------------------------------------------
    #           update_position
    # -----------------------------------------------------------------------------

    def update_position(self, utc: datetime | (np.datetime64 | float)) -> np.ndarray:
        """
        Overloads the methods from :class:`~.PlatformLocation`. Updates the position of the satellite to the requested time.
        The geocentric ECEF/ITRF position of the satellite can be retrieved with attribute :attr:`~position`. It is
        strongly recommended that time is always increasing in reasonable step sizes. Various orbit propagators will not
        step backwards in time  and may fail if the time step is unreasonably large.

        Parameters
        -----------
        utc : datetime
            The time at which to calculate the new position.

        """
        ut = skretrieval.time.ut_to_datetime(
            utc
        )  # The satellite code is based upon datetime, so ensure we havethe right time units
        self.update_eci_position(ut)
        return self.position

    # ------------------------------------------------------------------------------
    #           update_velocity
    # ------------------------------------------------------------------------------

    def update_velocity(
        self, utc: datetime | (np.datetime64 | float)
    ) -> np.ndarray | None:
        """
        Overloads the method from :class:`~.PlatformLocation`. Updates the velocity of the satellite to the requested time.
        The geocentric ECEF/ITRF velocity of the satellite can be retrieved with attribute :attr:`~velocity`. It is
        strongly recommended that time is always increasing in reasonable step sizes. Various orbit propagators will not
        step backwards in time  and may fail if the time step is unreasonably large.

        Parameters
        -----------
        utc : datetime
            The time at which to calculate the new velocity.
        """

        ut = skretrieval.time.ut_to_datetime(utc)
        self.update_eci_position(ut)
        return self.velocity

    # ------------------------------------------------------------------------------
    #           update_orientation
    # ------------------------------------------------------------------------------

    def update_orientation(
        self, utc: datetime | np.datetime64  # noqa: ARG002
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray] | None:
        """
        Returns a platform orientation that has no modification.
        """
        return (
            np.array((1.0, 0.0, 0.0)),
            np.array((0.0, 1.0, 0.0)),
            np.array((0.0, 0.0, 1.0)),
        )

    # -----------------------------------------------------------------------------
    #           update_eci_position
    # -----------------------------------------------------------------------------
    @abstractmethod
    def update_eci_position(self, utc: datetime):
        """
        All satellite classes must implement update_eci_position. This purely abstract method  will  update the internal attributes
        self._m_ecilocation, self._m_ecivelocity and self._m_time

        Parameters
        ----------
        utc : datetime
            Updates the ECI eciposition and ECI ecivelocity of the satellite to this Coordinated Universal Time
        """

        msg = "SatelliteBase::update_eci_position, Do not call base class method"
        raise Exception(msg)

    # ------------------------------------------------------------------------------
    #           period
    # ------------------------------------------------------------------------------
    @abstractmethod
    def period(self) -> timedelta:
        """
        Purely abstract method that returns the orbital period of the satellite in seconds.  For low earth orbit spacecraft
        this only needs to be accurate to ~0.1 seconds.

        Returns
        -------
        timedelta
            The period of the satellite in seconds
        """
        msg = "SatelliteBase::period, Do not call base class method"
        raise Exception(msg)

    @abstractmethod
    def eccentricity(self) -> float:
        """
        Purely abstract method that returns the eccentricty of the satellite orbit. This is used in the base class
        when locating the next crossing of teh ascending node and only requires moderate precision.

        Returns
        -------
        float
            The eccentricty of the satellite orbit in seconds
        """
        msg = "SatelliteBase::period, Do not call base class method"
        raise Exception(msg)

    # -----------------------------------------------------------------------------
    #           eciposition
    # -----------------------------------------------------------------------------
    def eciposition(self) -> np.ndarray:
        """
        The ECI cartesian position of the satellite in meters after the last call to :meth:`~.update_eci_position`.

        Returns
        -------
        np.ndarray[3]
            The three element array specifying ECI location as (X,Y,Z) meters
        """
        return self._m_ecilocation.copy()

    # ------------------------------------------------------------------------------
    #           ecivelocity
    # ------------------------------------------------------------------------------
    def ecivelocity(self) -> np.ndarray:
        """
        The ECI cartesian velocity of the satellite in meters per second

        Returns
        -------
        np.ndarray[3]
            The three element array specifying ECI velocity as (X,Y,Z) meters per second
        """
        return self._m_ecivelocity.copy()

    # -----------------------------------------------------------------------------
    #           _set_current_state
    # -----------------------------------------------------------------------------
    def _set_current_state(self, ecipos: np.ndarray, ecivel: np.ndarray, t: datetime):
        """
        Sets the ECI position, ECI velocity and current time of the satellite. This is typically
        called by child satellite implementation classes at the end of a call to method *update_eci_position*

        Parameters
        ----------
        ecipos : np.ndarray[3]
            The three element (X,Y,Z) array specifying ECI position in meters .
        ecivel : np.ndarray[3]
            The three element (X,Y,Z) array specifying ECI velocity in meters/second
        t  : datetime.datetime
            The current UTC time of the eciposition and ecivelocity.

        """
        self._m_ecilocation[
            :
        ] = ecipos  # Copy the values to ensure we have 3 element arrays
        self._m_ecivelocity[
            :
        ] = ecivel  # Copy the values to ensure we have 3 element arrays
        self._m_time = t

    # ------------------------------------------------------------------------------
    #      nxSatelliteBase::ZComponent
    # ------------------------------------------------------------------------------
    def _zcomponent(self, deltamjd: float) -> float:
        """
        Return the Z component of the satellite at a time deltamjd from
        the current epoch.  This is often used to find the ascending node crossing
        of the satellite.

        Parameters
        -----------
        deltamjd:float
            The delta time in seconds

        Returns
        -------
        float
            The Z component of the satellite ECI position. Negative means it is below the equator. Positive means it is above the equator.
        """

        Tnow = self._temptime + timedelta(
            seconds=deltamjd
        )  # Set up a new self.m_time to get satellite eciposition.
        self.update_eci_position(
            Tnow
        )  # Get the satellite eciposition atthis self.m_time.
        return self._m_ecilocation[2]  # and return the Z component.

    # ----------------------------------------------------------------------------
    #     nxSatelliteBase::equator_crossing
    # --------------------------------------------------------------------------
    def equator_crossing(self, utc: datetime) -> datetime:
        """
        Determines the last equator crossing before or equal to Tnow.

        Parameters
        ----------
        Tnow: datetime.datetime
            Find the equator crossing before (or at) this time

        Returns
        -------
        datetime.datetime
            The time at which this satellite object crosses the equator
        """

        self._temptime = utc
        ecc = self.eccentricity()
        frac = (
            0.2 if ecc < 0.1 else 0.01
        )  # small eccentricty we can do big steps. Large eccentricity do small steps
        deltat = (frac * self.period()).total_seconds()
        positiveT = (utc - self._temptime).total_seconds()
        z = -1
        while (
            z < 0.0
        ):  # ----  Find when Z component is positive before or equal to now.
            z = self._zcomponent(positiveT)
            if z < 0.0:
                positiveT = positiveT - deltat

        negativeT = (
            positiveT - deltat
        )  # ---- Find when Z component is negative before positiveT.
        z = 1.0
        while z >= 0.0:
            z = self._zcomponent(negativeT)
            if z >= 0.0:
                negativeT = negativeT - deltat

        crosstime = scipy.optimize.brentq(
            self._zcomponent, negativeT, positiveT, xtol=1.0e-12
        )
        positiveT = self._temptime + timedelta(seconds=crosstime)
        self.update_eci_position(positiveT)
        return positiveT

    # -----------------------------------------------------------------------------
    #           set_orbit_number_from_last_equator_crossing
    # -----------------------------------------------------------------------------
    def set_orbit_number_from_last_equator_crossing(
        self, orbitnumber: int, utc: datetime
    ):
        """
        Sets the orbit number and calculates the start time/ascending node of the orbit.

        Parameters
        ----------
        orbitnumber: int
            The orbit number at time Tnow
        Tnow: datetime.datetime
            The current time.
        """

        self._m_orbit_number_at_start_time = orbitnumber
        self._m_start_time_of_orbit = self.equator_crossing(utc)

    # ----------------------------------------------------------------------------
    #        nxSatelliteBase::orbit_number      2004-11-23*/
    # --------------------------------------------------------------------------

    def orbit_number(self, utc: datetime) -> tuple[int, datetime]:
        """
         Calculates the orbit number at time Tnow and returns the start time
         of that orbit.

        Parameters
        ----------
        utc : datetime
            The Coordinated Universal Time at which to calculate the orbit number

        Returns
        -------
        (int,datetime)
            Returns a two element tuple specifiying the orbit number and the time of the last equator crossing
        """
        Start = self.equator_crossing(utc)
        DeltaT = Start - self._m_start_time_of_orbit
        norbits = int(DeltaT / self.period() + 0.5)
        return (norbits + self._m_orbit_number_at_start_time, Start)



from __future__ import annotations

import abc

import numpy as np
from scipy.linalg import block_diag

from skretrieval.core.radianceformat import RadianceBase
from skretrieval.retrieval import RetrievalTarget
from skretrieval.retrieval.statevector import StateVector


class GenericTarget(RetrievalTarget):
    def measurement_vector(self, l1_data: RadianceBase):
        result = self._internal_measurement_vector(l1_data)

        if "jacobian" in result:
            result["jacobian"] = self.map_K(result["jacobian"])
        return result

    @abc.abstractmethod
    def _internal_measurement_vector(self, l1_data: RadianceBase):
        pass

    def state_vector(self):
        vec = []
        for state_element in self._state_vector.state_elements:
            if state_element.enabled:
                vec.append(state_element.state())

        return self._map_bounded_to_internal(np.concatenate(vec))

    def lower_bound(self):
        vec = []
        for state_element in self._state_vector.state_elements:
            if state_element.enabled:
                vec.append(state_element.lower_bound())

        return np.concatenate(vec)

    def upper_bound(self):
        vec = []
        for state_element in self._state_vector.state_elements:
            if state_element.enabled:
                vec.append(state_element.upper_bound())

        return np.concatenate(vec)

    def update_state(self, x: np.ndarray):
        rescaled_x = self._map_internal_to_bounded(x)

        for state_element, state_slice in zip(
            self._state_vector.state_elements, self._state_slices
        ):
            if state_element.enabled:
                state_element.update_state(rescaled_x[state_slice])

    def apriori_state(self) -> np.array:
        vec = []
        for state_element in self._state_vector.state_elements:
            if state_element.enabled:
                vec.append(state_element.apriori_state())
        return self._map_bounded_to_internal(np.concatenate(vec))

    def inverse_apriori_covariance(self):
        inv_covar = []
        for state_element in self._state_vector.state_elements:
            if state_element.enabled:
                inv_covar.append(state_element.inverse_apriori_covariance())

        return self._map_inv_Sa_by_dinternal(
            self.state_vector(), block_diag(*inv_covar)
        )

    def _map_bounded_to_internal(self, x: np.array) -> np.array:
        """
        Maps the bounded (user) state vector to the internal object

        Parameters
        ----------
        x : np.array
            Bounded state vector

        Returns
        -------
        np.array
            Internal state vector
        """
        if not self._rescale_state_elements:
            return x
        lb = self.lower_bound()
        ub = self.upper_bound()
        internal_x = np.zeros_like(x)

        no_map = (lb == -np.inf) & (ub == np.inf)

        both_bounds = (lb != -np.inf) & (ub != np.inf)

        internal_x[no_map] = x[no_map]

        internal_x[both_bounds] = np.arcsin(
            2 * (x[both_bounds] - lb[both_bounds]) / (ub[both_bounds] - lb[both_bounds])
            - 1
        )

        return internal_x

    def _map_internal_to_bounded(self, x: np.array) -> np.array:
        """
        Maps the internal state vector to the bounded state vector

        Parameters
        ----------
        x : np.array
            Internal state vector

        Returns
        -------
        np.array
            Bounded state vector
        """
        if not self._rescale_state_elements:
            return x
        lb = self.lower_bound()
        ub = self.upper_bound()
        bounded_x = np.zeros_like(x)

        no_map = (lb == -np.inf) & (ub == np.inf)

        both_bounds = (lb != -np.inf) & (ub != np.inf)

        bounded_x[no_map] = x[no_map]

        bounded_x[both_bounds] = (
            lb[both_bounds]
            + (np.sin(x[both_bounds]) + 1) * (ub[both_bounds] - lb[both_bounds]) / 2
        )

        return bounded_x

    def map_K(self, K: np.ndarray) -> np.ndarray:
        """
        Maps the bounded K (from SASKTRAN basically) to K for the internal variables

        Parameters
        ----------
        K : np.ndarray
            Jacobian matrix for the bounded state vector

        Returns
        -------
        np.ndarray
            Jacobian matrix for the internal state vector
        """
        if not self._rescale_state_elements:
            return K
        x = self.state_vector()
        lb = self.lower_bound()
        ub = self.upper_bound()
        mapping = np.zeros_like(x)

        no_map = (lb == -np.inf) & (ub == np.inf)

        both_bounds = (lb != -np.inf) & (ub != np.inf)

        mapping[no_map] = 1
        mapping[both_bounds] = (
            np.cos(x[both_bounds]) * (ub[both_bounds] - lb[both_bounds]) / 2
        )

        return K @ np.diag(mapping)

    def _map_inv_Sa_by_dinternal(self, x: np.array, inv_Sa: np.ndarray) -> np.ndarray:
        """
        Maps the inverse covariance for the bounded variables to the inverse covariance on the
        internal retrieval variables

        Parameters
        ----------
        x : np.array
            INTERNAL state vector
        inv_Sa : np.ndarray
            Bounded inverse covariance

        Returns
        -------
        np.ndarray
            Internal inverse covariance
        """
        if not self._rescale_state_elements:
            return inv_Sa
        lb = self.lower_bound()
        ub = self.upper_bound()
        xb = self._map_internal_to_bounded(x)
        mapping = np.zeros_like(x)
        no_map = (lb == -np.inf) & (ub == np.inf)

        both_bounds = (lb != -np.inf) & (ub != np.inf)

        mapping[no_map] = 1

        mapping[both_bounds] = 2 / (
            (ub[both_bounds] - lb[both_bounds])
            * np.sqrt(
                1
                - (
                    1
                    - (
                        2
                        * (xb[both_bounds] - lb[both_bounds])
                        / (ub[both_bounds] - lb[both_bounds])
                    )
                )
                ** 2
            )
        )

        return np.diag(1 / mapping) @ inv_Sa @ np.diag(1 / mapping)

    def __init__(self, state_vector: StateVector, rescale_state_elements: bool = False):
        """
        Implements a generic abstract base target class that is composed of a StateVector.  Derived classes of this
        type are responsible for implementing _internal_measurement_vector which computes the measurement vector.
        All of the other functionality, updating the state, apriori parameters, are handled by the state vector elements
        and this class.

        Parameters
        ----------
        state_vector: StateVector
            The state vector
        rescale_state_elements: bool
            If true, then the state vector respace is internally scaled so that the lower_bound and upper_bound of each
            state vector element is enforced.  This is useful if the minimization method does not allow for bounds,
            such as Rodgers or scipy LM
        """
        self._state_vector = state_vector
        self._rescale_state_elements = rescale_state_elements

        # Construct slices that map the full state vector to each individual state vector element
        self._state_slices = []
        cur_idx = 0
        for state_element in self._state_vector.state_elements:
            if state_element.enabled:
                n = len(state_element.state())
                self._state_slices.append(slice(cur_idx, cur_idx + n))
                cur_idx += n
            else:
                self._state_slices.append(None)



from __future__ import annotations

from .datetime64 import (
    datetime64_to_datetime,
    datetime64_to_timestamp,
    ut_to_datetime,
    ut_to_datetime64,
)
from .eci import eciteme_to_itrf, gmst, itrf_to_eciteme
from .juliandate import ut_to_jd
from .mjd import mjd_to_ut, ut_to_mjd, utc_to_astropy



import unittest
import numpy as np
from skretrieval.platforms.platform import Platform


class SkretrievalPlatformTests(unittest.TestCase):
    def test_instrument_mounting(self):
        """
        Tests the instrument mounting in the platform control frame
        """
        platform = Platform()
        vicf = np.array(((1, 0, 0), (0, 1, 0), (0, 0, 1))).transpose()
        utc = np.datetime64("2020-10-14T15:00:00.000000")
        platform.platform_pointing.set_platform_location(
            latlonheightandt=(0, 0, 0, utc)
        )

        platform.platform_pointing.mount_instrument_on_platform(60, 30.0, 60.0)
        vgeo = platform.platform_pointing.convert_icf_to_gcf(vicf)
        vecef = platform.platform_pointing.convert_icf_to_ecef(vicf)
        # print('rotation = (60,30,0)')
        # print('Geographic Unit vectors')
        # print(vgeo)
        # print('ECEF Unit vectors')
        # print(vecef)
        self.assertAlmostEqual(vgeo[0, 0], 0.75, 8)
        self.assertAlmostEqual(vgeo[0, 1], -0.625, 8)
        self.assertAlmostEqual(vgeo[0, 2], 0.21650635, 8)
        self.assertAlmostEqual(vgeo[1, 0], 0.4330127, 8)
        self.assertAlmostEqual(vgeo[1, 1], 0.21650635, 8)
        self.assertAlmostEqual(vgeo[1, 2], -0.875, 8)
        self.assertAlmostEqual(vgeo[2, 0], 0.5, 8)
        self.assertAlmostEqual(vgeo[2, 1], 0.75, 8)
        self.assertAlmostEqual(vgeo[2, 2], 0.4330127, 8)

        self.assertAlmostEqual(vgeo[0, 0], vecef[1, 0], 8)
        self.assertAlmostEqual(vgeo[1, 0], vecef[2, 0], 8)
        self.assertAlmostEqual(vgeo[2, 0], vecef[0, 0], 8)
        self.assertAlmostEqual(vgeo[0, 1], vecef[1, 1], 8)
        self.assertAlmostEqual(vgeo[1, 1], vecef[2, 1], 8)
        self.assertAlmostEqual(vgeo[2, 1], vecef[0, 1], 8)
        self.assertAlmostEqual(vgeo[0, 2], vecef[1, 2], 8)
        self.assertAlmostEqual(vgeo[1, 2], vecef[2, 2], 8)
        self.assertAlmostEqual(vgeo[2, 2], vecef[0, 2], 8)

    # ------------------------------------------------------------------------------
    #               def test_instrument_rotation_matrix(self):
    # ------------------------------------------------------------------------------
    def test_instrument_rotation_matrix(self):
        """
        Tests the rotation of the instrument in the instrument control frame.
        """
        platform = Platform()
        vicf = np.array(((1, 0, 0), (0, 1, 0), (0, 0, 1))).transpose()
        utc = np.datetime64("2020-10-14T15:00:00.000000")
        platform.platform_pointing.set_platform_location(
            latlonheightandt=(0, 0, 0, utc)
        )

        platform.platform_pointing.rotate_instrument_in_icf(30, 60.0, 0)
        vgeo = platform.platform_pointing.convert_icf_to_gcf(vicf)
        vecef = platform.platform_pointing.convert_icf_to_ecef(vicf)
        # print('Geographic Unit vectors')
        # print(vgeo)
        # print('ECEF Unit vectors')
        # print(vecef)
        # print('Done')
        self.assertAlmostEqual(vgeo[0, 0], 0.25, 8)
        self.assertAlmostEqual(vgeo[0, 1], -0.8660254, 8)
        self.assertAlmostEqual(vgeo[0, 2], -0.4330127, 8)
        self.assertAlmostEqual(vgeo[1, 0], 0.4330127, 8)
        self.assertAlmostEqual(vgeo[1, 1], 0.5, 8)
        self.assertAlmostEqual(vgeo[1, 2], -0.75, 8)
        self.assertAlmostEqual(vgeo[2, 0], 0.8660254, 8)
        self.assertAlmostEqual(vgeo[2, 1], 0.0, 8)
        self.assertAlmostEqual(vgeo[2, 2], 0.5, 8)

        self.assertAlmostEqual(vgeo[0, 0], vecef[1, 0], 8)
        self.assertAlmostEqual(vgeo[1, 0], vecef[2, 0], 8)
        self.assertAlmostEqual(vgeo[2, 0], vecef[0, 0], 8)
        self.assertAlmostEqual(vgeo[0, 1], vecef[1, 1], 8)
        self.assertAlmostEqual(vgeo[1, 1], vecef[2, 1], 8)
        self.assertAlmostEqual(vgeo[2, 1], vecef[0, 1], 8)
        self.assertAlmostEqual(vgeo[0, 2], vecef[1, 2], 8)
        self.assertAlmostEqual(vgeo[1, 2], vecef[2, 2], 8)
        self.assertAlmostEqual(vgeo[2, 2], vecef[0, 2], 8)


if __name__ == "__main__":
    tests = SkretrievalPlatformTests()
    unittest.main()



from __future__ import annotations

from collections.abc import Iterable
from copy import copy

import numpy as np
import xarray as xr
from sasktran import Geometry, LineOfSight

import skretrieval.core.radianceformat as radianceformat
from skretrieval.core import OpticalGeometry
from skretrieval.core.lineshape import LineShape
from skretrieval.legacy.core.sensor import Sensor
from skretrieval.legacy.core.sensor.pixel import Pixel


class SpectrographPixelArray(Sensor):
    """
    A spectrograph created from an array of Pixels
    """

    def __init__(
        self,
        wavelength_nm: np.array,
        pixel_shape: LineShape,
        vert_fov: LineShape,
        horiz_fov: LineShape,
    ):
        """

        Parameters
        ----------
        wavelength_nm : np.array
            Central wavelengths for each pixel
        pixel_shape: LineShape
            Wavelength line shape
        vert_fov: LineShape
            Vertical field of view
        horiz_fov: LineShape
            Horizontal field of view
        """
        self._pixels = []

        if isinstance(pixel_shape, Iterable):
            for w, p_shape in zip(wavelength_nm, pixel_shape):
                self._pixels.append(Pixel(w, p_shape, vert_fov, horiz_fov))
        else:
            for w in wavelength_nm:
                self._pixels.append(Pixel(w, pixel_shape, vert_fov, horiz_fov))

        self.wavelength_nm = wavelength_nm

    def measurement_geometry(self, optical_geometry: OpticalGeometry):
        return [
            LineOfSight(
                optical_geometry.mjd,
                optical_geometry.observer,
                optical_geometry.look_vector,
            )
        ]

    def model_radiance(
        self,
        optical_geometry: OpticalGeometry,
        model_wavel_nm: np.array,
        model_geometry: Geometry,
        radiance: np.array,
        wf=None,
    ):
        data = xr.concat(
            [
                p.model_radiance(
                    optical_geometry, model_wavel_nm, model_geometry, radiance, wf
                ).data
                for p in self._pixels
            ],
            dim="wavelength",
            data_vars="minimal",
        )

        return radianceformat.RadianceGridded(data)

    @staticmethod
    def radiance_format():
        return radianceformat.RadianceGridded

    def measurement_wavelengths(self):
        return self.wavelength_nm

    def required_wavelengths(self, res_nm: float) -> np.array:
        """
        Recommended wavelengths for high resolution calculations

        Parameters
        ----------
        res_nm: float
            Resolution in nm

        Returns
        -------
        np.array
        """
        lower_bounds = np.zeros(len(self._pixels))
        upper_bounds = np.zeros(len(self._pixels))

        for i, p in enumerate(self._pixels):
            lower_bounds[i], upper_bounds[i] = p._pixel_shape.bounds()

            upper_bounds[i] += p._wavelength_nm
            lower_bounds[i] += p._wavelength_nm

        bounding_sets = _set_join(lower_bounds, upper_bounds)

        wavel = np.concatenate(
            [np.arange(set[0], set[1], res_nm) for set in bounding_sets]
        )

        if len(wavel) == 0:
            wavel = np.array(self.wavelength_nm)

        return wavel


class Spectrograph(SpectrographPixelArray):
    """
    Same functionality as Spectrogaph, but coded to not consist of individual pixels for speed
    """

    def __init__(
        self,
        wavelength_nm: np.array,
        pixel_shape: LineShape,
        vert_fov: LineShape,
        horiz_fov: LineShape,
    ):
        """

        Parameters
        ----------
        wavelength_nm : np.array
            Central wavelengths for each pixel
        pixel_shape: LineShape
            Wavelength line shape
        vert_fov: LineShape
            Vertical field of view
        horiz_fov: LineShape
            Horizontal field of view
        """
        super().__init__(wavelength_nm, pixel_shape, vert_fov, horiz_fov)

        self._cached_wavel_interp = None
        self._cached_wavel_interp_wavel = None

    def _construct_interpolators(
        self, model_geometry, optical_geometry, model_wavel_nm
    ):
        los_interp = self._pixels[0]._construct_los_interpolator(
            model_geometry, optical_geometry
        )

        if not np.array_equal(model_wavel_nm, self._cached_wavel_interp_wavel):
            wavel_interp = []
            for p in self._pixels:
                wavel_interp.append(
                    p._construct_wavelength_interpolator(model_wavel_nm)
                )

            wavel_interp = np.vstack(wavel_interp)
            self._cached_wavel_interp = wavel_interp
            self._cached_wavel_interp_wavel = copy(model_wavel_nm)

        return self._cached_wavel_interp, los_interp

    def model_radiance(
        self,
        optical_geometry: OpticalGeometry,
        model_wavel_nm: np.array,
        model_geometry: Geometry,
        radiance: np.array,
        wf=None,
    ):
        wavel_interp, los_interp = self._construct_interpolators(
            model_geometry, optical_geometry, model_wavel_nm
        )

        modelled_radiance = np.einsum(
            "ij,jk...,kl", wavel_interp, radiance, los_interp, optimize="optimal"
        )

        data = xr.Dataset(
            {
                "radiance": (["wavelength", "los"], modelled_radiance),
                "mjd": (["los"], [optical_geometry.mjd]),
                "los_vectors": (
                    ["los", "xyz"],
                    optical_geometry.look_vector.reshape((1, 3)),
                ),
                "observer_position": (
                    ["los", "xyz"],
                    optical_geometry.observer.reshape((1, 3)),
                ),
            },
            coords={
                "wavelength": self.measurement_wavelengths(),
                "xyz": ["x", "y", "z"],
            },
        )

        if wf is not None:
            modelled_wf = np.einsum(
                "ij,jkl,km->iml", wavel_interp, wf, los_interp, optimize="optimal"
            )

            data["wf"] = (["wavelength", "los", "perturbation"], modelled_wf)

        return radianceformat.RadianceGridded(data)


def _set_join(lower_bounds, upper_bounds):
    final_sets = [[lower_bounds[0], upper_bounds[0]]]

    for lower, upper in zip(lower_bounds, upper_bounds):
        new_set = True
        for set in final_sets:
            if _in_set(set, lower) and not _in_set(set, upper):
                set[1] = upper
                new_set = False
                break
            elif not _in_set(set, lower) and _in_set(set, upper):  # noqa: RET508
                set[0] = lower
                new_set = False
                break
            elif _in_set(set, lower) and _in_set(set, upper):
                new_set = False
        if new_set:
            final_sets.append([lower, upper])
    return final_sets


def _in_set(set, val):
    if val >= set[0] and val <= set[1]:
        return True
    return None



from __future__ import annotations

import json
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any

import numpy as np
import skyfield
import skyfield.timelib
import tudatjson  # This is created in the ARG copy of Tudat Bundle, see tudatBundle\tudatApplications\python_json_dll\python_tudatjson_interface

from .satellitebase import SatelliteBase
from .satellitekepler import SatelliteKepler

jsontemplate = r"""
{
  "initialEpoch": 0,
  "finalEpoch": 86400,
  "globalFrameOrientation": "J2000",
  "spice": {
    "useStandardKernels": true,
    "preloadEphemeris": false
  },
  "bodies": {
    "Sun": {
      "useDefaultSettings": true
    },
    "Earth": {
      "useDefaultSettings": true,
      "gravityField": {
            "type": "sphericalHarmonic",
            "model": "ggm02c"
      }
    },
    "Moon": {
      "useDefaultSettings": true
    },
    "custom_satellite": {
      "initialState": {
        "type": "cartesian",
        "x": 1732339.13760243,
        "y": -6656846.40425629,
        "z": 723621.98106622,
        "vx": -775.84161121,
        "vy": -1026.88295639,
        "vz": -7488.07085716
      },
      "mass": 400,
      "referenceArea": 4,
      "atmosphere": {
            "type": "nrlmsise00"
       },
      "aerodynamics": {
        "forceCoefficients": [ 8.75, 0, 0 ]
      },
      "radiationPressure": {
        "Sun": {
          "radiationPressureCoefficient": 1.2,
          "occultingBodies": [ "Earth" ]
        }
      }
    }
  },
  "propagators": [
    {
      "integratedStateType": "translational",
      "centralBodies": [
        "Earth"
      ],
      "bodiesToPropagate": [
        "custom_satellite"
      ],
      "accelerations": {
        "custom_satellite": {
          "Earth": [
            {
              "type": "sphericalHarmonicGravity",
              "maximumDegree": 20,
              "maximumOrder": 20
            },
            {
              "type": "aerodynamic"
            }
          ],
          "Sun": [
            {
              "type": "pointMassGravity"
            },
            {
              "type": "cannonBallRadiationPressure"
            }
          ],
          "Moon": [
            {
              "type": "pointMassGravity"
            }
          ]
        }
      }
    }
  ],
  "integrator": {
    "type": "rungeKutta4",
    "stepSize": 10
  },
  "export": [
    {
      "file": "temporary_stateHistory.txt",
      "variables": [
        {
          "type": "state"
        }
      ]
    }
  ],
  "options": {
    "fullSettingsFile": "temporary_fullSettings.json"
  }
}
"""


# -----------------------------------------------------------------------------
#           StateCache
# -----------------------------------------------------------------------------
class StateCache:
    """
    Holds a cache of satellite state
    """

    def __init__(self):
        self.cache_starttime: float = None  # start time of first value in cache
        self.cache_endtime: float = None  # end time of last value in cache.
        self.entry_starttime: float = None  # start time of first value in current entry
        self.entry_endtime: float = None  # end time of last value in current entry.
        self.timeoffsets: np.ndarray = (
            None  # The array of time offsets in seconds since J2000.
        )
        self.positions: np.ndarray = None  # The array of platform_ecef_positions in the current simulation window
        self.velocities: np.ndarray = (
            None  # The array of velocities in the current simulation window
        )
        self.ts: skyfield.timelib.Timescale = (
            skyfield.api.load.timescale()
        )  # Load in the skyfield timescales
        self.j2000: skyfield.timelib.Time = self.ts.tt(
            2000, 1, 1, 12, 0, 0
        )  # Keep the time of J2000 cached
        self.kepler: SatelliteKepler = SatelliteKepler()

    # -----------------------------------------------------------------------------
    #           in_cache
    # -----------------------------------------------------------------------------

    def in_cache(self, tt_j2000: float) -> bool:
        return (tt_j2000 >= self.cache_starttime) and (tt_j2000 < self.cache_endtime)

    # -----------------------------------------------------------------------------
    #           is_valid
    # -----------------------------------------------------------------------------

    def is_valid(self) -> bool:
        return self.timeoffsets is not None

    # -----------------------------------------------------------------------------
    #           update_cache
    # -----------------------------------------------------------------------------
    def update_cache(self, tudatresults: np.ndarray):
        self.timeoffsets = tudatresults[:, 0]  # seconds since J2000.
        self.positions = tudatresults[:, 1:4]
        self.velocities = tudatresults[:, 4:7]
        self.cache_starttime = self.timeoffsets[0]
        self.cache_endtime = self.timeoffsets[-1]
        if self.cache_endtime < self.cache_starttime:
            np.flip(self.timeoffsets)
            np.flipud(self.positions)
            np.flipud(self.velocities)
            self.cache_starttime = self.timeoffsets[0]
            self.cache_endtime = self.timeoffsets[-1]
        self.entry_starttime = 1.0
        self.entry_endtime = 0.0

    # -----------------------------------------------------------------------------
    #           utc_to_tt_j2000
    # -----------------------------------------------------------------------------
    def datetime_to_ttj2000(self, mjd: datetime) -> float:
        t = self.ts.observation_policy.utc(
            mjd.year,
            mjd.month,
            mjd.day,
            mjd.hour,
            mjd.minute,
            mjd.second + mjd.microsecond / 1.0e6,
        )
        return (t.tt - self.j2000.tt) * 86400.0

    # -----------------------------------------------------------------------------
    #           ttj2000_to_datetime
    # -----------------------------------------------------------------------------
    def ttj2000_to_datetime(self, tt_j2000) -> datetime:
        t = self.ts.tt_jd(tt_j2000 / 86400.0 + self.j2000.tt)
        return t.utc_datetime()

    # -----------------------------------------------------------------------------
    #           update_entry
    # -----------------------------------------------------------------------------

    def update_entry(self, tt_j2000: float):
        ok = (tt_j2000 >= self.entry_starttime) and (tt_j2000 < self.entry_endtime)
        if not ok:
            index = np.searchsorted(self.timeoffsets, tt_j2000) - 1
            self.entry_starttime = self.timeoffsets[index]
            self.entry_endtime = self.timeoffsets[index + 1]
            r = self.positions[index, :]
            v = self.velocities[index, :]
            mjd0 = self.ttj2000_to_datetime(self.entry_starttime)
            self.kepler.from_state_vector(mjd0, 1, r, v)
            ok = True
        return ok

    # -----------------------------------------------------------------------------
    #           interpolate_current_entry
    # -----------------------------------------------------------------------------

    def interpolate_current_entry(
        self, tt_j2000: float
    ) -> tuple[np.ndarray, np.ndarray]:
        mjd = self.ttj2000_to_datetime(tt_j2000)
        self.kepler.update_eci_position(mjd)
        r = self.kepler.eciposition()
        v = self.kepler.ecivelocity()
        return (r, v)

    # -----------------------------------------------------------------------------
    #           interpolate_position_and_velocity
    # -----------------------------------------------------------------------------

    def interpolate_position_and_velocity(
        self, tt_j2000: float
    ) -> tuple[np.ndarray, np.ndarray]:
        ok = self.update_entry(tt_j2000)
        if ok:
            r, v = self.interpolate_current_entry(tt_j2000)
        else:
            r = None
            v = None
        return (r, v)


# ------------------------------------------------------------------------------
#           class SatelliteTudat
# ------------------------------------------------------------------------------
class SatelliteTudat(SatelliteBase):
    def __init__(self):
        super().__init__()
        self.cache: StateCache = StateCache()
        self.simulation_window: float = 86400.0  # Get the length of the simulation window in seconds. default is 1 day
        self.jsonobject: dict[str, Any] = json.loads(jsontemplate)

    # ---------------------------------------------------------------------------
    #              SatelliteKepler::orbital_period
    # ---------------------------------------------------------------------------

    def period(self) -> timedelta:
        """
        Return the orbital period of this orbit. Uses keplers third law.

        Returns
        -------
        datetime.timedelta
            The orbital period.
        """
        return self.cache.kepler.period()

    # -----------------------------------------------------------------------------
    #           calculate_cache
    # -----------------------------------------------------------------------------

    def calculate_cache(self, start_j2000, end_j2000, r: np.ndarray, v: np.ndarray):
        self.jsonobject["initialEpoch"] = start_j2000
        self.jsonobject["finalEpoch"] = end_j2000
        state = self.jsonobject["bodies"]["custom_satellite"]["initialState"]
        state["type"] = "cartesian"
        state["x"] = r[0]
        state["y"] = r[1]
        state["z"] = r[2]
        state["vx"] = v[0]
        state["vy"] = v[1]
        state["vz"] = v[2]
        jsonstring = json.dumps(self.jsonobject)
        self.jsonobject = tudatjson.execute_json(jsonstring)
        datafile = self.jsonobject["export"][0]["file"]
        settingsfile = self.jsonobject["options"]["fullSettingsFile"]
        results = np.loadtxt(datafile)
        Path.unlink(datafile)
        Path.unlink(settingsfile)
        self.cache.update_cache(results)

    # ---------------------------------------------------------------------------
    #      SatelliteKepler::update_eci_position
    # ---------------------------------------------------------------------------

    def update_eci_position(self, mjd: datetime):
        if (self._m_time is None) or (mjd != self._m_time):
            tt_j2000 = self.cache.datetime_to_ttj2000(mjd)
            ok = self.cache.in_cache(tt_j2000)  # this time is not in the current cache
            while not ok:  # so
                if self.cache.is_valid():  # If we have a valid cache
                    if (
                        tt_j2000 >= self.cache.cache_endtime
                    ):  # then if we are going forwards in time
                        startdt = (
                            self.cache.cache_endtime
                        )  # start at the end of the current cache
                        enddt = startdt + self.simulation_window
                        r = self.cache.positions[-1, :]
                        v = self.cache.velocities[-1, :]
                    else:  # otherwise we are going backwards in time
                        startdt = self.cache.cache_starttime
                        enddt = self.cache.cache_starttime - self.simulation_window
                        r = self.cache.positions[0, :]
                        v = self.cache.velocities[0, :]
                else:
                    msg = "You must initialize the tudat satellite propagator before trying to update the eci position"
                    raise Exception(msg)
                self.calculate_cache(startdt, enddt, r, v)
                ok = self.cache.in_cache(tt_j2000)
            self._m_time = mjd
            r, v = self.cache.interpolate_position_and_velocity(tt_j2000)
            self._m_location = r.copy()
            self._m_velocity = v.copy()

    # -----------------------------------------------------------------------------
    #           from_state_vector
    # -----------------------------------------------------------------------------

    def from_state_vector(
        self,
        mjd: datetime,
        orbitnumber: int,  # noqa: ARG002
        r: np.ndarray,
        v: np.ndarray,
    ):
        startdt = self.cache.datetime_to_ttj2000(mjd)
        enddt = startdt + self.simulation_window
        self.calculate_cache(startdt, enddt, r, v)
        self.update_eci_position(mjd)



from __future__ import annotations

from datetime import datetime

import astropy.coordinates
import astropy.units as units
import numpy as np
from astropy.time import Time

astropy.coordinates.solar_system_ephemeris.set(
    "jpl"
)  # Ensure the JPL ephemerides are used by default


# -----------------------------------------------------------------------------
#           planetary_body_itrf
# -----------------------------------------------------------------------------
def solsys_body_vector_itrf(utc: datetime, body: str) -> np.ndarray:
    """
    Fetches the ITRF geocentric vector from the center of the Earth to the solar system body at the requested time.

    Parameters
    ----------
    utc : datetime, np.datetime64, str
        The UTC time at which the position of the body is required
    body: str
        The name of the solar system body, e.g. "moon", "sun"

    Returns
    -------
    np.ndarray(3): The geocentric positon of the requested body as a 3 element column vector. Expressed in meters.
    """

    t = Time(utc)
    coord = astropy.coordinates.get_body(body, t, ephemeris="jpl")
    itrs = astropy.coordinates.ITRS(representation_type="cartesian")
    itrscoord = coord.transform_to(itrs)
    return np.array(
        (
            itrscoord.x.to(units.meter).value,
            itrscoord.y.to(units.meter).value,
            itrscoord.z.to(units.meter).value,
        )
    )


def star_unitvector_itrf(utc: datetime, body: str) -> np.ndarray:
    """
    Fetches the ITRF geocentric unit vector from the center of the Earth to the desired star at the requested time.
    This currently does not account for **proper motion** of the star.

    Parameters
    ----------
    utc : datetime, np.datetime64, str
        The UTC time at which the position of the star is required
    body: str
        The name of the star, e.g. "Betelgeuse"

    Returns
    -------
    np.ndarray(3): The geocentric unit vector toward the requested star as a 3 element column vector. Dimensionless Unit vector.
    """

    t = Time(utc)
    star = astropy.coordinates.SkyCoord.from_name(
        body
    )  # **** DOES NOT SEEM TO INCLUDE proper motion corrections
    itrs = astropy.coordinates.ITRS(representation_type="cartesian", obstime=t)
    itrscoord = star.transform_to(itrs)
    return np.array((itrscoord.x.value, itrscoord.y.value, itrscoord.z.value))



import unittest
from datetime import datetime, timedelta
from math import radians
import numpy as np
from skretrieval.platforms.satellite import (
    SatelliteBase,
    SatelliteKepler,
    SatelliteSGP4,
    SatelliteSunSync,
    SatelliteMolniya,
    SatelliteSimpleGeostationary,
)
from skretrieval.platforms.satellite.gibbs import gibbs
from skretrieval.platforms.satellite.coe import coe_from_state_vector


# -----------------------------------------------------------------------------
#           SkimpySatelliteTests
# -----------------------------------------------------------------------------
class SkretrievalSatelliteTests(unittest.TestCase):
    def setUp(self):
        self._numtimes = 200
        self._locations = np.zeros((3, self._numtimes))
        self._times = np.zeros([self._numtimes], dtype="datetime64[us]")
        self._utc = datetime(2019, 7, 26, hour=20, minute=15, second=00)
        dt = timedelta(minutes=1.0)
        for i in range(self._numtimes):  # for each time step
            tnow = self._utc + i * dt
            self._times[i] = tnow  # save the time of the step

    # -----------------------------------------------------------------------------
    #           _propagate_satellite
    # -----------------------------------------------------------------------------
    def _propagate_satellite(self, sat: SatelliteBase) -> np.ndarray:
        for i in range(self._numtimes):
            position = sat.update_position(
                self._times[i]
            )  # get the geographic geocentrix position of each satellite
            self._locations[:, i] = sat.lat_lon_height
            # print('{:3d} {:18.13f} {:18.13f} {:23.13f}'.format(i,self._locations[0,i], self._locations[1,i],self._locations[2,i] ))
        return self._locations

    def test_sunsync_twolinelements(self):
        utc = datetime.fromisoformat("2023-06-23 16:00:00.000000")
        t = datetime.fromisoformat("2023-06-23 16:10:00.000000")
        kepler = SatelliteKepler()
        sunsync = SatelliteSunSync(
            utc,
            orbittype="sgp4",
            period_from_altitude=450000.0,
            localtime_of_ascending_node_hours=13.5,
        )  # get a sun sync satellite, this is only accurate to a couple of seconds of time as it uses Kepler two line elements to initialize an SGP8 orbit.
        teq = sunsync.equator_crossing(t)  # get the time of the equator crossing
        satpos = sunsync.update_position(teq)  # get its position at this time
        ecipos = sunsync.eciposition()
        ecivel = sunsync.ecivelocity()
        kepler.from_state_vector(
            teq, ecipos, ecivel, 1000
        )  # get a kepler orbit that matches the SGP4
        kpos = kepler.update_position(teq)
        kecipos = kepler.eciposition()
        kecivel = kepler.ecivelocity()
        lines = kepler.make_two_line_elements()
        sgp4 = SatelliteSGP4(lines)
        sgp4pos = sgp4.update_position(teq)
        spos = sgp4.eciposition()
        svel = sgp4.ecivelocity()
        ds = sgp4pos - satpos
        s = np.linalg.norm(ds)
        self.assertLessEqual(s, 15500.0)

    # ------------------------------------------------------------------------------
    #           test_gibbs
    # ------------------------------------------------------------------------------
    def test_gibbs(self):
        r1 = np.array([-294320.0, 4265100.0, 5986700.0])
        r2 = np.array([-1365400.0, 3637600.0, 6346800.0])
        r3 = np.array([-2940300.0, 2473700.0, 6555800.0])
        v2 = gibbs(
            r1, r2, r3
        )  # Use Gibbs to calculate satellite velocity at the middle point of 3 points
        self.assertAlmostEqual(v2[0], -6217.6009496, 3)
        self.assertAlmostEqual(v2[1], -4012.3770592, 3)
        self.assertAlmostEqual(v2[2], 1599.15006807, 3)

        coe = coe_from_state_vector(
            r2, v2
        )  # use the satellite position and velocity to calculate the classical orbital elements
        self.assertAlmostEqual(coe.H * 1.0e-6, 56193.04806713033, 3)
        self.assertAlmostEqual(coe.RA, 0.6981716759913223, 8)
        self.assertAlmostEqual(coe.TA, 0.8707351512007719, 8)
        self.assertAlmostEqual(coe.W, 0.5255060496425946, 8)
        self.assertAlmostEqual(coe.a, 8002140.855890051, 2)
        self.assertAlmostEqual(coe.e, 0.1001592330355506, 9)
        self.assertAlmostEqual(coe.i, 1.0472156172639657, 8)

        kepler = SatelliteKepler()
        kepler.from_state_vector(self._utc, r2, v2)
        p = kepler.period()
        self.assertAlmostEqual(p.total_seconds(), 7123.940246, 3)

    # ------------------------------------------------------------------------------
    #           test_three_points
    # ------------------------------------------------------------------------------
    def test_three_points(self):
        r1 = np.array([-294320.0, 4265100.0, 5986700.0])
        r2 = np.array([-1365400.0, 3637600.0, 6346800.0])
        r3 = np.array([-2940300.0, 2473700.0, 6555800.0])

        kepler = SatelliteKepler()
        kepler.from_three_positions(self._utc, r1, r2, r3)
        p = kepler.period()
        self.assertAlmostEqual(p.total_seconds(), 7123.940246, 3)

    # -----------------------------------------------------------------------------
    #           test_SGP4
    # -----------------------------------------------------------------------------

    def test_SGP4(self):  # create our different satellites::
        line1 = "1 26702U 01007A   19206.65122582  .00000348  00000-0  26798-4 0  9995"  # two line elements that will be used for the SGP4 satellite
        line2 = "2 26702  97.5720 223.8268 0009640 316.2599  43.7864 15.07871693  7426"
        sgp4 = SatelliteSGP4(twolines=[line1, line2])

        for nloop in range(
            4
        ):  # Do 4 loops to ensure we can step backwards in time without issue
            positions = self._propagate_satellite(sgp4)
            self.assertAlmostEqual(
                positions[0, 99], 2.3180298075593, 2
            )  # Verify Latitude
            self.assertAlmostEqual(
                positions[1, 99], -48.0400704025521, 2
            )  # Verify Longitude
            self.assertAlmostEqual(
                positions[2, 99], 541249.5916230629664, 2
            )  # Verify Altitude
            self.assertAlmostEqual(
                positions[0, 10], 26.960762147970545, 2
            )  # Verify Latitude
            self.assertAlmostEqual(
                positions[1, 10], -29.33926117654872, 2
            )  # Verify Longitude
            self.assertAlmostEqual(
                positions[2, 10], 544393.0853459131, 2
            )  # Verify Altitude

    # -----------------------------------------------------------------------------
    #           test_Kepler
    # -----------------------------------------------------------------------------
    def test_Kepler(self):
        kepler = SatelliteKepler(
            self._utc,
            period_from_altitude=600000.0,
            inclination_radians=radians(97.0),
            longitude_of_ascending_node_degrees=82.0,
            eccentricity=0.05,
        )
        for nloop in range(4):
            positions = self._propagate_satellite(kepler)
            self.assertAlmostEqual(
                positions[0, 99], 9.514217229999177, 2
            )  # Verify Latitude
            self.assertAlmostEqual(
                positions[1, 99], 56.010676386422446, 2
            )  # Verify Longitude
            self.assertAlmostEqual(
                positions[2, 99], 256026.54192426076, 2
            )  # Verify Altitude
        # print("Done Kepler")

    def test_Molniya_with_kepler(self):
        molniya = SatelliteMolniya(
            self._utc, orbittype="kepler", longitude_of_ascending_node_degrees=-124.0
        )
        for nloop in range(4):
            positions = self._propagate_satellite(molniya)
            self.assertAlmostEqual(
                positions[0, 99], 61.99186405390857, 2
            )  # Verify Latitude
            self.assertAlmostEqual(
                positions[1, 99], -78.6699517266697, 2
            )  # Verify Longitude
            self.assertAlmostEqual(
                positions[2, 99], 38227950.71962145, 2
            )  # Verify Altitude
        # print("Done molniya with kepler")

    def test_Molniya_with_sgp4(self):
        molniya = SatelliteMolniya(
            self._utc, orbittype="sgp4", longitude_of_ascending_node_degrees=-124.0
        )
        for nloop in range(4):
            positions = self._propagate_satellite(molniya)
            self.assertAlmostEqual(
                positions[0, 99], 61.99170529136819, 2
            )  # Verify Latitude
            self.assertAlmostEqual(
                positions[1, 99], -78.72355534491894, 2
            )  # Verify Longitude
            self.assertAlmostEqual(
                positions[2, 99], 38206706.45955145, 2
            )  # Verify Altitude
        # print("Done molniya with sgp4")

    def test_sunsync_with_kepler(self):
        sunsync = SatelliteSunSync(
            self._utc,
            orbittype="kepler",
            period_from_altitude=600000.0,
            localtime_of_ascending_node_hours=18.25,
        )
        for nloop in range(4):
            positions = self._propagate_satellite(sunsync)
            self.assertAlmostEqual(
                positions[0, 99], 8.583229967118568, 2
            )  # Verify Latitude
            self.assertAlmostEqual(
                positions[1, 99], -55.993522413900365, 2
            )  # Verify Longitude
            self.assertAlmostEqual(
                positions[2, 99], 600470.0139338467, 2
            )  # Verify Altitude
        # print("Done sun sync with kepler")

    def test_sunsync_with_sgp4(self):
        sunsync = SatelliteSunSync(
            self._utc,
            orbittype="sgp4",
            period_from_altitude=600000.0,
            localtime_of_ascending_node_hours=18.25,
        )
        for nloop in range(4):
            positions = self._propagate_satellite(sunsync)
            self.assertAlmostEqual(
                positions[0, 99], 8.23901273452463, 2
            )  # Verify Latitude
            self.assertAlmostEqual(
                positions[1, 99], -55.87883826556556, 2
            )  # Verify Longitude
            self.assertAlmostEqual(
                positions[2, 99], 602335.9354597726, 2
            )  # Verify Altitude
        # print("Done sun sync with sgp4")

    def test_simple_geostationary(self):
        geostat = SatelliteSimpleGeostationary(-80.0)
        for nloop in range(4):
            positions = self._propagate_satellite(geostat)
            self.assertAlmostEqual(
                positions[0, 99], 0.0000000000000, 2
            )  # Verify Latitude
            self.assertAlmostEqual(
                positions[1, 99], -80.0000000000000, 2
            )  # Verify Longitude
            self.assertAlmostEqual(
                positions[2, 99], 35785863.0000000000000, 2
            )  # Verify Altitude



from __future__ import annotations

import numpy as np
import sasktran as sk

from skretrieval.core.sasktranformat import SASKTRANRadiance


def test_sasktran_legacy_format_vector():
    tanalts_km = np.arange(10, 50, 1)

    # First recreate our geometry and atmosphere classes
    geometry = sk.VerticalImage()
    geometry.from_sza_saa(
        sza=60,
        saa=60,
        lat=0,
        lon=0,
        tanalts_km=tanalts_km,
        mjd=54372,
        locallook=0,
        satalt_km=600,
        refalt_km=20,
    )

    atmosphere = sk.Atmosphere()

    atmosphere["ozone"] = sk.Species(sk.O3OSIRISRes(), sk.Labow())
    atmosphere["air"] = sk.Species(sk.Rayleigh(), sk.MSIS90())

    # And now make the engine
    engine = sk.EngineHR(geometry=geometry, atmosphere=atmosphere)

    # Choose some wavelengths to do the calculation at
    engine.wavelengths = [340, 600]
    engine.polarization = "vector"

    atmosphere.wf_species = "ozone"

    # And do the calculation
    radiance = engine.calculate_radiance(
        full_stokes_vector=True, output_format="xarray"
    )

    _ = SASKTRANRadiance.from_sasktran_legacy_xr(radiance)


def test_sasktran_legacy_format_scalar():
    tanalts_km = np.arange(10, 50, 1)

    # First recreate our geometry and atmosphere classes
    geometry = sk.VerticalImage()
    geometry.from_sza_saa(
        sza=60,
        saa=60,
        lat=0,
        lon=0,
        tanalts_km=tanalts_km,
        mjd=54372,
        locallook=0,
        satalt_km=600,
        refalt_km=20,
    )

    atmosphere = sk.Atmosphere()

    atmosphere["ozone"] = sk.Species(sk.O3OSIRISRes(), sk.Labow())
    atmosphere["air"] = sk.Species(sk.Rayleigh(), sk.MSIS90())

    # And now make the engine
    engine = sk.EngineHR(geometry=geometry, atmosphere=atmosphere)

    # Choose some wavelengths to do the calculation at
    engine.wavelengths = [340, 600]
    # engine.polarization = 'vector'

    atmosphere.wf_species = "ozone"

    # And do the calculation
    radiance = engine.calculate_radiance(
        full_stokes_vector=False, output_format="xarray"
    )

    _ = SASKTRANRadiance.from_sasktran_legacy_xr(radiance)


def test_sasktran2_format():
    pass



from __future__ import annotations

import logging
import math
import numbers
from datetime import datetime, timezone
from typing import Sequence

import numpy as np

from .mjd import mjd_to_datetime64


# ------------------------------------------------------------------------------
#           def datetime64_to_timestamp_new( utc ):
# ------------------------------------------------------------------------------
def datetime64_to_timestamp(utc):
    """
    Converts a scalar or array of numpy.datetime64 to floating point timestamp (seconds since 1970). The conversion includes
    the fraction of seconds.

    Parameters
    ----------
    utc : numpy.datetime64 or Array[numpy.datetime64]
        A time specified with a numpy datetime64 object. Explicit time zones are
        not currently supported. Only single scalar values are supported in this function.

    Returns
    -------
    float
        The time expressed as a timestamp value. Number of seconds since Jan 1, 1970.
    """

    if type(utc) is np.ndarray:
        if utc.dtype.kind == "M":  # if we have an array of datetime64
            tstamp = np.zeros_like(utc, dtype=np.float64)  # create the output array
            with np.nditer(
                [utc, tstamp],
                flags=["refs_ok", "buffered"],  # create the numpy iterator object
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # and convert each elemnet
                    t = (
                        np.datetime64(a, "us")
                        .astype(datetime)
                        .replace(tzinfo=timezone.utc)
                        .timestamp()
                    )  # from datetime64 to a timestamp
                    b[...] = t
                tstamp = it.operands[1]

        # Handle an numpy array of MJD values, datetime64, datetime objects or mjd numbers
        elif (
            (utc.dtype.kind == "O") and (len(utc) > 0) and (type(utc[0]) is datetime)
        ):  # otherwiseif we have a numpy array of  datetime objects
            tstamp = np.zeros_like(utc, dtype=np.float64)  # create the output array
            with np.nditer(
                [utc, tstamp],
                flags=["refs_ok", "buffered"],  # create the nupy iterator object
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # and convert each elemnet
                    t = a.replace(
                        tzinfo=timezone.utc
                    ).timestamp()  # from datetime to mjd
                    b[...] = t
                tstamp = it.operands[1]
        elif utc.dtype.kind == "U":  # otherwise if the array is strings
            tstamp = np.zeros_like(
                utc, dtype=np.float64
            )  # then make an array to hold the answer
            with np.nditer(
                [utc, tstamp],
                flags=["refs_ok", "buffered"],
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # For each element in the array
                    t = datetime.fromisoformat(str(a))  # convert the string to datetime
                    b[...] = t.replace(
                        tzinfo=timezone.utc
                    ).timestamp()  # and from datetime to mjd
                tstamp = it.operands[1]
        else:  # otherwise we have an array of something else, I assume its array of numbers
            tstamp = np.array(
                utc, dtype=np.float64
            )  # so try and convert the array of numbers to floating point MJD.

    elif isinstance(utc, str):  # if we have a single string
        tstamp = datetime.fromisoformat(
            utc
        ).timestamp()  # then convert the string to datetime

    elif isinstance(utc, Sequence):  # if we have a list or tuple but not numpy array
        newutc = np.array(utc)  # then convert the array to numpy array
        tstamp = datetime64_to_timestamp(newutc)  # and convert the numpy array to mjd

    else:  # Its not a numpy array, its not a sequence, so assume its a scalar
        if type(utc) is datetime:  # if its a datetime object
            tstamp = utc.replace(
                tzinfo=timezone.utc
            ).timestamp()  # to mjd as a floating point
        elif type(utc) is np.datetime64:  # if we have a single numpy.datetime64
            tstamp = (
                np.datetime64(utc, "us")
                .astype(datetime)
                .replace(tzinfo=timezone.utc)
                .timestamp()
            )  # then convert it
        else:
            logging.warning(
                "utc_to_timestamp, unsupported data type. Cannot robustly convert to timestamp. I will try and convert it to a float",
                extra={"utc": utc},
            )
            tstamp = float(utc)  # then assume its already an MJD
    return tstamp


# ------------------------------------------------------------------------------
#           datetime64_to_datetime
# ------------------------------------------------------------------------------
def datetime64_to_datetime(usertime: np.datetime64) -> datetime:
    """
    Converts a single numpy.datetime64 to a datetime.datetime

    Parameters
    ----------
    usertime : numpy.datetime64
        A time specified with a numpy datetime64 object. Explicit time zones are
        not currently supported. Only single scalar values are supported in this function.

    Returns
    -------
    datetime.datetime
        The time expressed as a regular python datetime.datetime.
    """
    return np.datetime64(usertime, "us").astype(datetime)


# ------------------------------------------------------------------------------
#           datetime_to_datetime64
# ------------------------------------------------------------------------------
def datetime_to_datetime64(usertime: np.datetime64) -> np.datetime64:
    """
    Converts a single datetime to a np.datetime64

    Parameters
    ----------
    usertime : datetime.datetime
        A time specified with a python datetime object. Explicit time zones are
        not currently supported. Only single scalar values are supported in this function.

    Returns
    -------
    datetime.datetime
        The time expressed as a regular python datetime.datetime.
    """
    return np.datetime64(usertime, "us")


# ------------------------------------------------------------------------------
#           ut_to_datetime64
# ------------------------------------------------------------------------------
def ut_to_datetime64(utc):
    """
    A convenience function that converts various arrays or scalar representations of UT to datetime64
    Scalar input values will return as a scalar datyetime64 while array, list or tuple input values are all returned
    as numpy arrays of utc in datetime64.

    Parameters
    ----------

    utc : scalar, list, tuple or array
        The input time which represents a coordinated universal time. It can be represented by
        (i) a string in a supported python datetime.isoformat
        (ii) a number. The number is assumed to represent a modified julian date.
        (iii) a datetime.datetime object.
        (iv) a numpy.datetime64 object.
        The *utc* object can be a scalar, list, tuple or numpy array. The same representation format must be used for
        all objects in the arrays and sequences.

    Returns
    -------
    scalar or numpy.ndarray of np.datetim64
        The time is returned as a scalar if the input was a scalar or as a numpy array for input sequences and arrays
    """
    d64 = None
    if type(utc) is np.ndarray:
        # Handle an numpy array of MJD values, datetime64, datetime objects or mjd numbers
        if utc.dtype.kind == "M":  # if we have a numpy array of datetime64 object
            d64 = utc  # then do nothing
        elif (
            (utc.dtype.kind == "O") and (len(utc) > 0) and (type(utc[0]) is datetime)
        ):  # otherwiseif we have a numpy array of  datetime objects
            d64 = np.zeros_like(utc, dtype="datetime64[us]")  # create the output array
            with np.nditer(
                [utc, d64], flags=["refs_ok"], op_flags=[["readwrite"], ["writeonly"]]
            ) as it:
                for a, b in it:  # and convert each elemnet
                    t = datetime_to_datetime64(
                        a.tolist()
                    )  # The object comes in as a zero shape object of datetime. But the toplist converts it back to a datetime
                    b[...] = t
                d64 = it.operands[1]
        elif utc.dtype.kind == "U":  # otherwise if the array is strings
            d64 = np.zeros_like(
                utc, dtype="datetime64[us]"
            )  # then make an array to hold the answer
            with np.nditer(
                [utc, d64],
                flags=["refs_ok", "buffered"],
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # For each element in the array
                    b[...] = np.datetime64(str(a), "us")  # and from datetime to mjd
                d64 = it.operands[1]
        else:  # otherwise we have an array of something else, I assume its array of numbers
            d64 = np.zeros_like(
                utc, dtype="datetime64[us]"
            )  # then make an array to hold the answer
            with np.nditer(
                [utc, d64],
                flags=["refs_ok", "buffered"],
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # For each element in the array
                    b[...] = mjd_to_datetime64(
                        float(a)
                    )  # atry and convert the MJD to datetime64
                d64 = it.operands[1]
    elif isinstance(utc, str):  # if we have a single string
        d64 = np.datetime64(utc, "us")  # then convert the string to datetime

    elif isinstance(utc, Sequence):  # if we have a list or tuple but not numpy array
        newutc = np.array(utc)  # then convert the array to numpy array
        d64 = ut_to_datetime64(newutc)  # and convert the numpy array to mjd
    else:  # Its not a numpy array, its not a sequence, so assume its a scalar
        if type(utc) is datetime:  # if its a datetime object
            d64 = datetime_to_datetime64(utc)  # to mjd as a floating point
        elif type(utc) is np.datetime64:  # if we have a single numpy.datetime64
            d64 = utc  # then convert it
        elif isinstance(utc, numbers.Number):  # if we have a number
            d64 = mjd_to_datetime64(float(utc))  # then assume its already an MJD
        else:  # otherwise
            d64 = math.nan
            msg = "unsupported data type {}. Cannot yet convert this to np.datetime64".format(
                str(type(utc))
            )
            raise ValueError(msg)
    return d64


# ------------------------------------------------------------------------------
#           ut_to_datetime64
# ------------------------------------------------------------------------------
def ut_to_datetime(
    utc: np.ndarray | (Sequence | (float | (str | (np.datetime64 | datetime)))),
) -> np.ndarray | (datetime | None):
    """
    A convenience function that converts various arrays or scalar representations of UT to datetime
    Scalar input values will return as a scalar datetime while array, list or tuple input values are all returned
    as numpy arrays of datetime objects

    Parameters
    ----------

    utc : scalar, list, tuple or array
        The input time which represents a coordinated universal time. It can be represented by
        (i) a string in a supported python datetime.isoformat
        (ii) a number. The number is assumed to represent a modified julian date.
        (iii) a datetime.datetime object.
        (iv) a numpy.datetime64 object.
        The *utc* object can be a scalar, list, tuple or numpy array. The same representation format must be used for
        all objects in the arrays and sequences.

    Returns
    -------
    scalar or numpy.ndarray of datetime
        The time is returned as a scalar if the input was a scalar or as a numpy array for input sequences and arrays
    """
    d64 = None
    if type(utc) is np.ndarray:
        # Handle an numpy array of MJD values, datetime64, datetime objects or mjd numbers
        if utc.dtype.kind == "O":  # we have a numpy array of  datetime objects
            d64 = utc  # then do nothing
            if (len(utc) > 0) and (type(utc[0]) is not datetime):
                msg = "ut_to_datetime does support conversion of arrays of python objects other than datetime"
                raise ValueError(msg)
        elif (
            utc.dtype.kind == "M"
        ):  # otherwiseif we have a numpy array of  datetime64 objects
            d64 = np.zeros_like(utc, dtype="O")  # create the output array
            with np.nditer(
                [utc, d64], flags=["refs_ok"], op_flags=[["readwrite"], ["writeonly"]]
            ) as it:
                for a, b in it:  # and convert each elemnet
                    t = datetime64_to_datetime(
                        a.tolist()
                    )  # The object comes in as a zero shape object of datetime. But the toplist converts it back to a datetime
                    b[...] = t
                d64 = it.operands[1]
        elif utc.dtype.kind == "U":  # otherwise if the array is strings
            d64 = np.zeros_like(utc, dtype="O")  # then make an array to hold the answer
            with np.nditer(
                [utc, d64],
                flags=["refs_ok", "buffered"],
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # For each element in the array
                    b[...] = datetime.fromisoformat(str(a))  # and from datetime to mjd
                d64 = it.operands[1]
        else:  # otherwise we have an array of something else, I assume its array of numbers
            d64 = np.zeros_like(utc, dtype="O")  # then make an array to hold the answer
            with np.nditer(
                [utc, d64],
                flags=["refs_ok", "buffered"],
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # For each element in the array
                    t = mjd_to_datetime64(float(a))  # Convert the mjd to datetime 64
                    b[...] = datetime64_to_datetime(t)  # and then convert to  datetime
                d64 = it.operands[1]
    elif isinstance(utc, str):  # if we have a single string
        d64 = datetime.fromisoformat(utc)  # then convert the string to datetime
    elif isinstance(utc, Sequence):  # if we have a list or tuple but not numpy array
        newutc = np.array(utc)  # then convert the array to numpy array
        d64 = ut_to_datetime(newutc)  # and convert the numpy array to mjd

    else:  # Its not a numpy array, its not a sequence, so assume its a scalar
        if type(utc) is datetime:  # if its a datetime object
            d64 = utc  # to mjd as a floating point
        elif type(utc) is np.datetime64:  # if we have a single numpy.datetime64
            d64 = datetime64_to_datetime(utc)  # then convert it
        elif isinstance(utc, numbers.Number):  # if we have a number
            t = mjd_to_datetime64(float(utc))  # then assume its already an MJD
            d64 = datetime64_to_datetime(t)
        else:  # otherwise
            msg = "unsupported data type {}. Cannot yet convert this to np.datetime64".format(
                str(type(utc))
            )
            raise ValueError(msg)
    return d64



from __future__ import annotations

import unittest

import numpy as np
from sasktran import Geometry, LineOfSight

from skretrieval.core import OpticalGeometry
from skretrieval.core.lineshape import Gaussian
from skretrieval.legacy.core.sensor.pixel import Pixel


class TestPixel(unittest.TestCase):
    def test_initialization(self):
        wavel_nm = 350

        line_shape = Gaussian(fwhm=1)

        _ = Pixel(wavel_nm, line_shape, Gaussian(fwhm=1), Gaussian(fwhm=1))

    def test_line_shape(self):
        wavel_nm = 350

        line_shape = Gaussian(fwhm=1)

        optical_geo = OpticalGeometry(
            observer=[0, 0, 1], look_vector=[0, 1, 0], local_up=[0, 0, 1], mjd=54372
        )

        pixel = Pixel(wavel_nm, line_shape, Gaussian(fwhm=1), Gaussian(fwhm=1))

        hires_wavel = pixel.required_wavelengths(0.01)

        model_geometry = Geometry()

        angles = np.linspace(-0.1, 0.1, 100)

        for a in angles:
            model_geometry.lines_of_sight.append(
                LineOfSight(
                    mjd=54372, observer=[0, 0, 1], look_vector=[0, np.cos(a), np.sin(a)]
                )
            )

        rad = np.zeros((len(hires_wavel), len(angles)))

        rad[300, 50] = 1

        _ = pixel.model_radiance(optical_geo, hires_wavel, model_geometry, rad)



from __future__ import annotations

import numpy as np
import sasktran as sk
import sasktran2 as sk2

from skretrieval.core import OpticalGeometry
from skretrieval.core.lineshape import Gaussian, Rectangle
from skretrieval.core.sasktranformat import SASKTRANRadiance
from skretrieval.core.sensor.spectrograph import Spectrograph
from skretrieval.legacy.util import convert_sasktran_legacy_geometry


def test_spectrograph():
    alt_grid = np.arange(0, 65001, 1000.0)

    tanalts_km = np.arange(10, 60, 2.0)

    geo = sk.VerticalImage()

    geo.from_sza_saa(
        sza=60,
        saa=60,
        lat=0,
        lon=0,
        tanalts_km=tanalts_km,
        mjd=54372,
        locallook=0,
        satalt_km=600,
        refalt_km=20,
    )

    new_geo = convert_sasktran_legacy_geometry(geo)

    model_geometry = sk2.Geometry1D(
        np.cos(np.deg2rad(60)),
        0,
        6372000,
        alt_grid,
        sk2.InterpolationMethod.LinearInterpolation,
        sk2.GeometryType.Spherical,
    )

    config = sk2.Config()

    engine = sk2.Engine(config, model_geometry, new_geo)

    atmosphere = sk2.Atmosphere(
        model_geometry, config, wavelengths_nm=np.arange(280, 800, 0.1)
    )

    sk2.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk2.constituent.Rayleigh()
    atmosphere["o3"] = sk2.climatology.mipas.constituent("O3", sk2.optical.O3DBM())

    radiance = SASKTRANRadiance.from_sasktran2(
        engine.calculate_radiance(atmosphere), collapse_scalar_stokes=True
    )
    radiance.data["look_vectors"] = (
        ["los", "xyz"],
        np.vstack([los.look_vector for los in geo.lines_of_sight]),
    )

    wavel_grid = np.arange(280, 800, 1.0)

    pixel_shape = [Gaussian(fwhm=1) for _ in wavel_grid]

    spectrograph = Spectrograph(
        wavel_grid, pixel_shape, Gaussian(fwhm=1), Rectangle(1, mode="constant")
    )

    optical_geo = OpticalGeometry(
        observer=geo.lines_of_sight[10].observer,
        look_vector=geo.lines_of_sight[10].look_vector,
        mjd=geo.lines_of_sight[10].mjd,
        local_up=geo.lines_of_sight[10].observer
        / np.linalg.norm(geo.lines_of_sight[10].observer),
    )

    _ = spectrograph.model_radiance(radiance=radiance, orientation=optical_geo)



from __future__ import annotations

import unittest

from skretrieval.core.platform.satellite import fake_satellite_scan


class TestFakeSatelliteScan(unittest.TestCase):
    def test_basic(self):
        _ = fake_satellite_scan()



from __future__ import annotations

import logging
import math
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..orientation_techniques import PointingAlgorithms
    from ..platform import Platform

import numpy as np
import scipy.optimize

from skretrieval.geodetic import geodetic
from skretrieval.time import datetime64_to_datetime

from ..platform.rotationmatrix import UnitVectors


# ------------------------------------------------------------------------------
#           ObserverBearingFunction
# ------------------------------------------------------------------------------
class ObserverBearingFunction:
    """
    An internal callable object that is used when calculating bearings of tangent points from the observer
    """

    def __init__(self, tanlat, tanlng, tanheight, observerheight, observer_bearing):
        self.target_observer_bearing = observer_bearing
        self._geo = geodetic()
        self._geo.from_lat_lon_alt(tanlat, tanlng, tanheight)
        self.tpnorth = -self._geo.local_south
        self.tpeast = -self._geo.local_west
        self.tplocation = self._geo.location
        self.obslocation = None
        self.observerheight = observerheight

        if self.target_observer_bearing > 180.0:
            self.target_observer_bearing -= 360.0

    def __call__(self, angledeg):
        angle = math.radians(
            angledeg
        )  # start with a nominal bearing at tangent point from observer to tangent point
        h = (
            math.cos(angle) * self.tpnorth + math.sin(angle) * self.tpeast
        )  # get the horizontal unit vector towards the observer from tan point
        points = self._geo.altitude_intercepts(self.observerheight, self.tplocation, h)
        self.obslocation = points[0]  # get the new observer location
        self._geo.from_xyz(self.obslocation)  # and set the location
        obsnorth = (
            -self._geo.local_south
        )  # get the local horizontal vectors at the observer location
        obseast = -self._geo.local_west
        xval = obsnorth.dot(h)  # get the northward component
        yval = obseast.dot(h)  # get the eastward componet
        bearing = math.degrees(
            math.atan2(yval, xval)
        )  # get the bearing of the tangent point from the observer
        offset = bearing - self.target_observer_bearing
        if offset < -180:
            offset += 360.0
        if offset > 180:
            offset -= 360.0
        # print('Angle deg {}, bearing at observer {}, offset {} ({},{}) at {}'.format( angledeg, bearing,  offset, self._geo.latitude, self._geo.longitude, self._geo.altitude))
        return offset


# ------------------------------------------------------------------------------
#           _technique_set_instrument_internal_orientation
# ------------------------------------------------------------------------------
def _technique_set_instrument_internal_orientation(
    platform: Platform, ut: np.datetime64, turntable_definition: np.ndarray
):
    platform.platform_pointing.rotate_instrument_in_icf(
        turntable_definition[0], turntable_definition[1], turntable_definition[2]
    )
    return True


# ------------------------------------------------------------------------------
#           _technique_set_platform_position_from_xyz
# ------------------------------------------------------------------------------
def _technique_set_platform_position_from_xyz(
    platform: Platform, ut: np.datetime64, position_definition: np.ndarray
):
    platform.platform_pointing.set_platform_location(
        xyzt=(
            position_definition[0],
            position_definition[1],
            position_definition[2],
            ut,
        )
    )
    return True


# ------------------------------------------------------------------------------
#           _technique_set_platform_position_from_llh
# ------------------------------------------------------------------------------
def _technique_set_platform_position_from_llh(
    platform: Platform, ut: np.datetime64, position_definition: np.ndarray
) -> bool:
    platform.platform_pointing.set_platform_location(
        latlonheightandt=(
            position_definition[0],
            position_definition[1],
            position_definition[2],
            ut,
        )
    )
    return True


# ------------------------------------------------------------------------------
#           _technique_set_platform_position_from_llh
# ------------------------------------------------------------------------------
def _technique_set_platform_position_from_observer_looking_at_llh(
    platform: Platform, ut: np.datetime64, position_definition: np.ndarray
) -> bool:
    latitude = position_definition[0]
    longitude = position_definition[1]
    height = position_definition[2]
    observer_bearing = position_definition[3]
    if observer_bearing > 180.0:
        observer_bearing -= 360.0
    observer_height = position_definition[4]
    ok = (
        (latitude > -91.0)
        and (latitude < 91.0)
        and (longitude > -361)
        and (longitude < 721)
        and (observer_bearing > -361)
        and (observer_bearing < 721)
        and (observer_height >= height)
    )
    if not ok:
        msg = "_technique_set_platform_position_from_observer_looking_at_llh parameters that were out of acceptable range. You might want to check you havethe correct parameter order"
        raise ValueError(msg)

    bearingfunc = ObserverBearingFunction(
        latitude, longitude, height, observer_height, observer_bearing
    )  # Initialize the observer bearing calculator function
    minval = 9999.0
    minangle = 9999.0
    for angle in range(
        -180, 190, 10
    ):  # Coarse step through azimuths at the tangent point
        val = math.fabs(bearingfunc(angle))  # to find the azimuths at the tangent point
        if (
            val < minval
        ):  # that are generate the proper bearing required at the observer
            minval = val  # the calculation is usually within 10 centimeters.
            minangle = angle

    roots = scipy.optimize.brentq(
        bearingfunc,
        minangle - 10,
        minangle + 10,
        xtol=1.0e-04,
        rtol=1.0e-06,
        full_output=True,
    )
    angle = roots[0]
    converged = roots[1].converged
    if converged:
        obs = bearingfunc.obslocation
        bearingfunc._geo.from_xyz(obs)
        latitude = (
            bearingfunc._geo.latitude
        )  # Get the latitude of the observer location
        longitude = (
            bearingfunc._geo.longitude
        )  # Get the longitude of the observer position
        platform.platform_pointing.set_platform_location(
            latlonheightandt=(latitude, longitude, observer_height, ut)
        )  # But get the height from the user defined value so it is bang on.
    if not converged:
        logging.warning(
            "technique_set_platform_position_from_observer_looking_at_llh, cannot find observer position that matches the bearing requiremenets"
        )
    return converged


# ------------------------------------------------------------------------------
#           _technique_set_platform_position_from_platform
# ------------------------------------------------------------------------------
def _technique_set_platform_position_from_platform(
    platform: Platform, ut: np.datetime64, position_definition: np.ndarray
) -> bool:
    utc = datetime64_to_datetime(ut)
    locator = platform.platform_locator
    ok = locator is not None
    if not ok:
        logging.warning(
            "You must add an instance of PlatformLocator to the Platform class if you wish to get positions data using positioning technique 'from_platform'"
        )
    else:
        position = platform.platform_locator.update_position(utc)
        ok = position is not None
        if ok:
            platform.platform_pointing.set_platform_location(
                xyzt=(position[0], position[1], position[2], ut)
            )
    return ok


# ------------------------------------------------------------------------------
#           from_tangent_altitudes
# ------------------------------------------------------------------------------
def _technique_set_look_vectors_from_tangent_altitude(
    pointing_algorithm: PointingAlgorithms,
    look_vector_definition: np.ndarray,
    roll_control: str,
) -> bool:
    platform = pointing_algorithm.platform
    observer = platform.platform_pointing.location()
    tangent_altitude = look_vector_definition[0]
    geographic_bearing_degrees = look_vector_definition[1]
    roll_angle = look_vector_definition[2] if look_vector_definition.size == 3 else 0.0
    return pointing_algorithm.set_limb_boresight_to_look_at_tangent_altitude(
        observer, tangent_altitude, geographic_bearing_degrees, roll_control, roll_angle
    )


# ------------------------------------------------------------------------------
#           set_limb_look_vectors_from_xyz
# ------------------------------------------------------------------------------
def _technique_set_limb_look_vectors_from_unit_xyz(
    pointing_algorithm: PointingAlgorithms,
    look_vector_definition: np.ndarray,
    roll_control: str,
) -> bool:
    observer = pointing_algorithm.platform.platform_pointing.location()
    look = look_vector_definition[0:3]
    roll_angle = look_vector_definition[3] if look_vector_definition.size == 4 else 0.0
    return pointing_algorithm.set_limb_boresight_from_lookvector(
        observer, look, roll_control, roll_angle
    )


# ------------------------------------------------------------------------------
#           set_nadir_look_vectors_from_xyz
# ------------------------------------------------------------------------------
def _technique_set_boresight_look_at_location_xyz(
    pointing_algorithm: PointingAlgorithms,
    look_vector_definition: np.ndarray,
    roll_control: str,
) -> bool:
    platform = pointing_algorithm.platform
    observer = platform.platform_pointing.location()
    target = look_vector_definition[0:3]
    roll_angle = look_vector_definition[3] if look_vector_definition.size == 4 else 0.0
    return pointing_algorithm.set_boresight_to_look_at_geocentric_location(
        observer, target, roll_control, roll_angle
    )


# ------------------------------------------------------------------------------
#           set_nadir_look_vectors_from_xyz
# ------------------------------------------------------------------------------
def _technique_set_boresight_look_at_location_llh(
    pointing_algorithm: PointingAlgorithms,
    look_vector_definition: np.ndarray,
    roll_control: str,
) -> bool:
    platform = pointing_algorithm.platform
    observer = platform.platform_pointing.location()
    latitude = look_vector_definition[0]
    longitude = look_vector_definition[1]
    height = look_vector_definition[2]
    roll_angle = look_vector_definition[3] if look_vector_definition.size == 4 else 0.0
    pointing_algorithm.geo.from_lat_lon_alt(latitude, longitude, height)
    target = pointing_algorithm.geo.location
    return pointing_algorithm.set_boresight_to_look_at_geocentric_location(
        observer, target, roll_control, roll_angle
    )


# ------------------------------------------------------------------------------
#           set_nadir_look_vectors_from_xyz
# ------------------------------------------------------------------------------
def _technique_set_boresight_pointing_from_unitvectors(
    pointing_algorithm: PointingAlgorithms,
    look_vector_definition: np.ndarray,
    roll_control: str,
) -> bool:
    xunit = look_vector_definition[0:3]
    zunit = look_vector_definition[3:6]
    yunit = np.cross(zunit, xunit)
    G = UnitVectors(vectors=(xunit, yunit, zunit))
    pointing_algorithm.set_boresight_pointing_from_unitvectors(G)
    return True


# ------------------------------------------------------------------------------
#           set_nadir_look_vectors_from_xyz
# ------------------------------------------------------------------------------
def _technique_set_observer_to_look_in_azi_elev(
    pointing_algorithm: PointingAlgorithms,
    look_vector_definition: np.ndarray,
    roll_control: str,
) -> bool:
    platform = pointing_algorithm.platform
    observer = platform.platform_pointing.location()
    azimuth = look_vector_definition[0]
    elevation = look_vector_definition[1]
    roll_angle = look_vector_definition[2] if look_vector_definition.size == 3 else 0.0
    return pointing_algorithm.set_observer_to_look_in_azi_elev(
        observer, elevation, azimuth, roll_control, roll_angle
    )


# ------------------------------------------------------------------------------
#           _technique_set_platform_pointing_from_platform
# ------------------------------------------------------------------------------
def _technique_set_platform_pointing_from_platform(
    pointing_algorithm: PointingAlgorithms,
    look_vector_definition: np.ndarray,
    roll_control: str,
) -> bool:
    platform = pointing_algorithm.platform
    utc = platform.platform_pointing.utc()
    locator = platform.platform_locator
    ok = locator is not None
    if not ok:
        logging.warning(
            "You must add an instance of PlatformLocator to the Platform class if you wish to get look vector data using pointing technique 'from_platform'"
        )
    else:
        vectors = platform.platform_locator.update_orientation(utc)
        ok = vectors is not None
        if ok:
            G = UnitVectors(vectors=vectors)
            pointing_algorithm.set_boresight_pointing_from_unitvectors(G)
    return ok


# ------------------------------------------------------------------------------
#           set_nadir_look_vectors_from_xyz
# ------------------------------------------------------------------------------
def _technique_set_boresight_look_at_location_orbitangle(
    pointing_algorithm: PointingAlgorithms,
    look_vector_definition: np.ndarray,
    roll_control: str,
) -> bool:
    platform = pointing_algorithm.platform
    observer = platform.platform_locator.position
    velocity = platform.platform_locator.velocity
    v = np.linalg.norm(velocity)
    ok = v > 0.0
    if ok:
        vunit = velocity / v
        pointing_algorithm.geo.from_xyz(observer)
        north = -pointing_algorithm.geo.local_south
        east = -pointing_algorithm.geo.local_west
        x = np.dot(north, vunit)
        y = np.dot(
            east, vunit
        )  # get the eastward component of the velocity at the observers location
        theta = math.degrees(math.atan2(y, x))
        tangent_altitude = look_vector_definition[0]
        orbitplane_bearing_degrees = look_vector_definition[1]
        geographic_bearing_degrees = theta + orbitplane_bearing_degrees
        roll_angle = (
            look_vector_definition[2] if look_vector_definition.size == 3 else 0.0
        )
        ok = pointing_algorithm.set_limb_boresight_to_look_at_tangent_altitude(
            observer,
            tangent_altitude,
            geographic_bearing_degrees,
            roll_control,
            roll_angle,
        )
    return ok


# ------------------------------------------------------------------------------
#           _technique_set_platform_position_from_platform
# ------------------------------------------------------------------------------
def _technique_set_icf_orientation_from_azi_elev(
    platform: Platform, azielevdata: np.ndarray
) -> bool:
    azimuth = azielevdata[0]
    elevation = azielevdata[1]
    roll = azielevdata[2] if azielevdata.size == 3 else 0
    platform.platform_pointing.rotate_instrument_in_icf(azimuth, elevation, roll)
    return True


# ------------------------------------------------------------------------------
#           _technique_set_platform_position_from_platform
# ------------------------------------------------------------------------------
def _technique_set_icf_look_from_xyz(platform: Platform, xyz_data: np.ndarray) -> bool:
    raise AssertionError()  # platform.platform_pointing.set_platform_location(xyzt=(position[0], position[1], position[2], ut))
    return True


# ------------------------------------------------------------------------------
#           _technique_set_platform_position_from_platform
# ------------------------------------------------------------------------------
def _technique_set_icf_orientation_no_operation(
    platform: Platform, xyz_data: np.ndarray
) -> bool:
    platform.platform_pointing.reset_icf_rotation_matrices()
    return True



from __future__ import annotations

import logging

import numpy as np
import sasktran as sk
import scipy as sp
from scipy import interpolate

from skretrieval.core import OpticalGeometry
from skretrieval.util import rotation_matrix


class OrbitalPlaneGrid:
    """
    Class which constructs a two dimensional grid out of a set of tangent
    points and look directions.  The core idea is that when the earths rotation
    is corrected for, the tangent points should form a plane.  A uniform in angle
    grid is constructed in this plane.  The approximation made is that over the
    course of one radiative transfer calculation, the actual grid is almost the
    same as this rotation corrected grid.

    The inputs to this function should be the low resolution values

    Parameters
    ----------
    tp : list or 2d numpy array
        Tangent points, should be convertible to a numpy array of shape(n,3)
    look : list or 2d numpy array
        Look directions, should be convertible to a numpy array of shape(n,3)
    mjd : List or numpy array
        Modified julian dates
    scanslices : list
        Each list element should be a range indicating which line of sights
        belong to which scan. e.g. scanslices[0] could be range(0,60) indicating
        that lines of sight 0-60 are in the first scan
    numinbetween : scalar, optional
        The ratio of scans / retrieval profiles if placementtype is set to uniform
    cutoff : scalar, optional
        Angle in degrees to cutoff of each side of the grid if placementype is
        set to uniform
    extend : scalar, optional
        Angle in degrees to extend each side of the grid if placementtype is set to uniform. Default 0. Do not use
        both extend and cutoff
    placementtype : string, optional
        If set to 'scan', retrieval profiles are placed at the average tp of
        every scan.  If set to uniform, a uniform grid is constructed between
        the first and last scan
    geoid_mode : string, optional
        The geoid to use when calculating latitude/longitude etc.

    """

    def __init__(
        self,
        opt_geo: list[OpticalGeometry],
        grid_altitudes,
        numinbetween=1.0,
        cutoff=0,
        extend=0,
        placementtype="scan",
        geoid_mode="wgs84",
    ):
        self._geoid_mode = geoid_mode
        self._numinbetween = numinbetween  # The number of scans / ret profiles
        self.cutoff = np.deg2rad(cutoff)
        self._extend = np.deg2rad(extend)
        self._numextend = 0

        # Find the tangent point of each measurements
        self._averagescantp = []
        self._averagescanlook = []
        self._averagescanmjd = []
        self._altitudes = grid_altitudes

        (
            self._averagescantp,
            self._averagescanlook,
            self._averagescanmjd,
        ) = self._tangent_points(opt_geo)

        self.averagescanangles = None
        self._xaxis = None
        self._yaxis = None
        self._numretprof = None
        self._angles = None
        self._retprofilemjd = None
        self._retprofileavglook = None
        self._retprofileloc = None

        self._placementtype = placementtype

        self._make_retrieval_profile_locations()

    def _make_retrieval_profile_locations(self):
        """
        Takes the measurement grid and constructs the retrieval grid.
        """
        rottp = self._correct_for_earths_rotation(
            self._averagescantp, self._averagescanmjd, self._averagescanmjd[0]
        )
        localangles = np.zeros(np.size(self._averagescanmjd))
        xaxis = rottp[0, :] / np.linalg.norm(rottp[0, :])
        locallook = rottp[np.min([10, np.size(self._averagescanmjd)])] - rottp[0, :]
        locallook /= np.linalg.norm(locallook)
        normal = np.cross(locallook, xaxis)
        normal /= np.linalg.norm(normal)
        yaxis = np.cross(xaxis, normal)
        yaxis /= np.linalg.norm(yaxis)
        twopiaddition = 0
        for idx in range(1, np.size(localangles)):
            localangles[idx] = (
                np.arctan2(yaxis.dot(rottp[idx, :]), xaxis.dot(rottp[idx, :]))
                + twopiaddition
            )
            if localangles[idx] < (localangles[idx - 1] - np.pi):
                twopiaddition += 2 * np.pi
                localangles[idx] += 2 * np.pi

        self.averagescanangles = localangles
        self._xaxis = xaxis
        self._yaxis = yaxis

        totalanglechange = np.max(localangles)
        averageanglechange = (
            totalanglechange / np.size(self._averagescanmjd) * self._numinbetween
        )
        possibleretangles = np.linspace(
            0, totalanglechange, int(np.floor(totalanglechange / averageanglechange))
        )

        if self._placementtype == "scan":
            possibleretangles = localangles
        elif self._placementtype == "uniform":
            possibleretangles = possibleretangles[
                (possibleretangles > self.cutoff)
                & (possibleretangles < (totalanglechange - self.cutoff))
            ]

            if self._extend > 0:
                delta_angle = possibleretangles[1] - possibleretangles[0]
                num_extend = int(np.floor(self._extend / delta_angle))
                self._numextend = num_extend

                lower_add = (
                    np.arange(num_extend, 0.5, -1) * delta_angle * -1
                    + possibleretangles[0]
                )
                upper_add = (
                    np.arange(1, num_extend + 0.5, 1) * delta_angle
                    + possibleretangles[-1]
                )

                possibleretangles = np.concatenate(
                    (lower_add, possibleretangles, upper_add)
                )
        else:
            msg = "Placement type is not one of ['scan', 'uniform']"
            raise ValueError(msg)

        self._retprofileloc = np.zeros((np.size(possibleretangles), 3))
        self._retprofileavglook = np.zeros((np.size(possibleretangles), 3))
        self._retprofilemjd = np.zeros(np.size(possibleretangles))
        self._numretprof = np.size(possibleretangles)
        self._angles = possibleretangles

        for idx in range(self._numretprof):
            self._retprofileloc[idx, :] = rotation_matrix(
                normal, -1 * possibleretangles[idx]
            ).dot(rottp[0, :])

        self._retprofilemjd = np.interp(
            possibleretangles, localangles, self._averagescanmjd
        )
        f = interpolate.interp1d(
            localangles, self._averagescanlook, axis=0, fill_value="extrapolate"
        )  # Extrapolate in case we are extending the profiles
        self._retprofileavglook = f(possibleretangles)

        self._retprofileloc = self._correct_for_earths_rotation(
            self._retprofileloc, -1 * self._retprofilemjd, -1 * self._averagescanmjd[0]
        )

    def get_local_plane(self, loc: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
        """
        Returns back the local plane at a specific location.

        Parameters
        ----------
        loc : numpy array
            Shape (3,) tangent point

        Returns
        -------
        angles : numpy array
            Angles in radians in the plane which correspond to the profile locations
            angles are measured with respect to loc
        angleidx : numpy array
            Indicies of the angles in the grid
        normalandreference : numpy array
            The normal and reference of the returned local plane, this is intended
            to be passed into ISKEngine with the 'normalandreference' option
            Result is shape(6,) with the first 3 elements being the normal
            and the last 3 being the reference
        """
        anglesbetween = np.zeros(self._numretprof)
        for idx in range(self._numretprof):
            anglesbetween[idx] = np.arccos(
                loc.dot(self._retprofileloc[idx, :])
                / np.linalg.norm(loc)
                / np.linalg.norm(self._retprofileloc[idx, :])
            )

        minidx = np.argmin(anglesbetween)
        # Use loc as the reference (xaxis), and calculate the normal from the
        # tangent points near
        reference = loc / np.linalg.norm(loc)
        # TODO: check that this gradient is good?
        avglook = (
            np.gradient(self._retprofileloc, axis=0)
            / np.linalg.norm(np.gradient(self._retprofileloc, axis=0), axis=1)[
                :, np.newaxis
            ]
        )[minidx]
        normal = np.cross(avglook, reference)
        normal /= np.linalg.norm(normal)
        yaxis = np.cross(reference, normal)
        yaxis /= np.linalg.norm(yaxis)

        # Rotate the profile locations by the earth's rotation to find the
        # angles in the tangent plane
        # rottp = correct_for_earths_rotation( self.retprofileloc, self.retprofilemjd, self.retprofilemjd[minidx])
        # NEW METHOD: Project the retrieval profile locations onto the plane
        rottp = self.project_onto_plane(self._retprofileloc, normal)
        allangles = np.zeros(self._numretprof)
        twopiaddition = 0
        for idx in range(self._numretprof):
            allangles[idx] = np.arctan2(
                yaxis.dot(rottp[idx, :]), reference.dot(rottp[idx, :])
            )
            if idx > 0 and allangles[idx] < (allangles[idx - 1] - np.pi):
                twopiaddition += 2 * np.pi
                allangles[idx] += 2 * np.pi

        # Have to be careful because we do not assume periodicity at the edges
        # TODO: make this user set
        maxanglediff = np.deg2rad(20)

        lowangleidx = ((allangles >= -1 * maxanglediff) & (allangles < 0)).nonzero()[0]
        lowangleidx = 0 if np.size(lowangleidx) == 0 else lowangleidx[0]
        highangleidx = (allangles[minidx:] > maxanglediff).nonzero()[0]
        if np.size(highangleidx) == 0:
            highangleidx = self._numretprof
        else:
            highangleidx = highangleidx[0] + minidx

        angleidx = np.arange(lowangleidx, highangleidx)
        if np.size(angleidx) == 0:
            angleidx = minidx

        localangles = allangles[angleidx]
        sortidx = np.argsort(localangles)
        localangles = localangles[sortidx]
        angleidx = angleidx[sortidx]

        # Return back what we need
        return localangles, angleidx, np.concatenate((normal, reference))

    def interpolate_onto_grid(
        self,
        prof_lon: np.ndarray,
        prof_lat: np.ndarray,
        prof_height: np.ndarray,
        prof: np.ndarray,
        do_height_interp=True,
    ) -> np.ndarray:
        """
        Converts a lat/lon prof that is defined on any grid to the retrieval grid
        Parameters
        ----------
        prof_lon:
            Longitude for the profile
        prof_lat:
            Latitude for the profile
        prof_height:
            Heights for the profile.  If the profile is only two dimensions this is not used
        prof:
            The profile. (lat, lon, height)
        do_height_interp:
            True if you want to do height interpolation, false otherwise.

        Returns
        -------
        (lat, lon, height) profile on the retrieval grid
        """
        retlat, retlon, retheight = self._lat_lon_alts(self._retprofileloc)
        retlon[retlon > 180] -= 360
        prof_lon = np.asarray(prof_lon)
        prof_lon[prof_lon > 180] -= 360
        retheight = self._altitudes

        # First get vertical profiles at the retrieval locations
        s = np.shape(prof)

        if len(s) > 1:
            # The profile is a 2d profile
            temp_prof = np.zeros((len(retlat), s[1]))
            # profile is (location,height), so loop over heights
            for idx in range(s[1]):
                temp_prof[:, idx] = self.distance_interpolate(
                    prof_lat, prof_lon, retlat, retlon, prof[:, idx]
                )

            if do_height_interp:
                # Now construct the retrieval profile by interpolating over height
                ret_prof = np.zeros((len(retlat), len(retheight)))
                for idx in range(len(retlat)):
                    ret_prof[idx, :] = np.interp(
                        retheight, prof_height, temp_prof[idx, :]
                    )

                return ret_prof
            return temp_prof

        return self.distance_interpolate(prof_lat, prof_lon, retlat, retlon, prof)

    def interpolate_from_grid(
        self,
        prof_lon: np.ndarray,
        prof_lat: np.ndarray,
        prof_height: np.ndarray,
        ret_prof: np.ndarray,
        do_height_interp=True,
    ) -> np.ndarray:
        """
        Converts a profile that is specified on the retrieval grid to a new lat/lon grid
        Parameters
        ----------
        prof_lon:
            Longitude to interpolate onto
        prof_lat:
            Latitude to interpolate onto
        prof_height:
            Heights to interpolate onto
        ret_prof:
            Profile that is specified on the retrieval grid
        do_height_interp:
            True if height interpolation should be done

        Returns
        -------
        Profile on the new grid
        """
        retlat, retlon, retheight = self._lat_lon_alts(self._retprofileloc)
        retlon[retlon > 180] -= 360
        prof_lon = np.asarray(prof_lon)
        prof_lon[prof_lon > 180] -= 360
        retheight = self._altitudes

        # First get vertical profiles at the retrieval locations
        s = np.shape(ret_prof)

        if len(s) > 1:
            # The profile is a 2d profile
            temp_prof = np.zeros((len(retlat), s[1]))
            # profile is (location,height), so loop over heights
            for idx in range(s[1]):
                temp_prof[:, idx] = self.distance_interpolate(
                    retlat, retlon, prof_lat, prof_lon, ret_prof[:, idx]
                )
            if do_height_interp:
                # Now construct the retrieval profile by interpolating over height
                meas_prof = np.zeros((len(prof_lat), len(prof_height)))
                for idx in range(len(prof_lat)):
                    meas_prof[idx, :] = np.interp(
                        prof_height, retheight, temp_prof[idx, :]
                    )

                return meas_prof
            return temp_prof

        return self.distance_interpolate(retlat, retlon, prof_lat, prof_lon, ret_prof)

    @staticmethod
    def distance_interpolate(
        from_lat: np.ndarray,
        from_lon: np.ndarray,
        to_lat: np.ndarray,
        to_lon: np.ndarray,
        profile: np.ndarray,
    ) -> np.ndarray:
        """
        Interpolates a profile in spherical distance.

        Parameters
        ----------
        from_lat : np.ndarray
            Array length (n) of latitudes in degrees for the grid to interpolate from
        from_lon : np.ndarray
            Array length (n) of longitudes in degrees for the grid to interpolate from
        to_lat : np.ndarray
            Array length (m) of latitudes in degrees for the grid to interpolate to
        to_lon : np.ndarray
            Array length (m) of longitudes in degrees for the grid to interpolate to
        profile : np.ndarray
            Profile of size (n) that is to be interpolated

        Returns
        -------
        np.ndarray
            Profile of size (m)
        """
        new_profile = np.zeros(len(to_lat))

        end_lon = from_lon[-1]
        end_lat = from_lat[-1]
        dist_grid_to_end = np.cos(np.deg2rad(end_lat)) * np.cos(
            np.deg2rad(from_lat)
        ) * np.cos(np.deg2rad(from_lon - end_lon)) + np.sin(
            np.deg2rad(from_lat)
        ) * np.sin(
            np.deg2rad(end_lat)
        )
        dist_grid_to_end[dist_grid_to_end > 1] = 1
        dist_grid_to_end[dist_grid_to_end < -1] = -1
        dist_grid_to_end = np.arccos(dist_grid_to_end)

        for idx, (lat, lon) in enumerate(zip(to_lat, to_lon)):
            # Calculate the distance from the this point to all old locations
            # assuming a spherical earth
            distance = np.arccos(
                np.cos(np.deg2rad(lat))
                * np.cos(np.deg2rad(from_lat))
                * np.cos(np.deg2rad(from_lon - lon))
                + np.sin(np.deg2rad(from_lat)) * np.sin(np.deg2rad(lat))
            )

            # Check if we are outside the grid
            if np.all(np.diff(distance) > 0):
                new_profile[idx] = profile[0]
                continue
            if np.all(np.diff(distance) < 0):
                new_profile[idx] = profile[-1]
                continue

            distance *= np.sign(np.gradient(distance))
            useful_idx = np.abs(distance) < np.pi / 2

            if not (
                np.all(np.diff(distance[useful_idx]) > 0)
                or np.all(np.diff(distance[useful_idx]) < 0)
            ):
                logging.warning(
                    "Distance between points is not monotonic, using the nearest point instead"
                )
                new_profile[idx] = sp.interpolate.interp1d(
                    distance, profile, kind="nearest"
                )(0.0)

            if distance[useful_idx][1] < distance[useful_idx][0]:
                new_profile[idx] = np.interp(
                    [0.0], distance[useful_idx][::-1], profile[useful_idx][::-1]
                )
            else:
                new_profile[idx] = np.interp(
                    [0.0], distance[useful_idx], profile[useful_idx]
                )

        return new_profile

    def _tangent_points(self, opt_geo: list[OpticalGeometry]):
        geo = sk.Geodetic(self._geoid_mode)

        tan_locs = []
        tan_looks = []
        tan_mjds = []

        for o in opt_geo:
            geo.from_tangent_point(o.observer, o.look_vector)

            tan_locs.append(geo.location)
            tan_looks.append(o.look_vector)
            tan_mjds.append(o.mjd)

        return np.asarray(tan_locs), np.asarray(tan_looks), np.asarray(tan_mjds)

    @staticmethod
    def _correct_for_earths_rotation(tp, mjd, refmjd=None):
        """
        Takes a set of tangent points, corresponding mjds and rotates them by
        the Earth's rotation.    If you take a set of satellite tangent points
        and mjd's and apply this function, the end result should be a plane

        Parameters
        ----------
        tp: np.ndarray
            (n, 3) of tangent points
        mjd: np.ndarray
            (n,) modified julian dates
        refmjd : float
            The reference mjd where no rotation is applied.  If set to None the mean MJD is used

        Returns
        -------
        2d numpy array
            Shape (n,3) of rotated tangent points
        """
        if refmjd is None:
            refmjd = np.nanmean(mjd)
        omega = 2 * np.pi

        mjddiff = mjd - refmjd
        rottp = np.zeros(np.shape(tp))

        for idx in range(np.size(mjddiff)):
            theta = omega * mjddiff[idx]
            rot_mat = np.array(
                [
                    [np.cos(theta), -1 * np.sin(theta), 0],
                    [np.sin(theta), np.cos(theta), 0],
                    [0, 0, 1],
                ]
            )

            rottp[idx, :] = rot_mat.dot(tp[idx, :])

        return rottp

    @staticmethod
    def project_onto_plane(tp: np.ndarray, normal: np.ndarray) -> np.ndarray:
        """
        Takes a list of tangent point vectors and returns their value projected onto a plane specified by normal.
        Note that the projected value returned has the same magnitude as the original input tangent point.  The only
        thing different is the direction
        Parameters
        ----------
        tp: np.ndarray
            (:, 3) array of tangent points
        normal: np.ndarray
            (3,) normal unit vector for the plane

        Returns
        -------
        projected: np.ndarray
            The tangent points projected into the plane, same size as tp
        """
        projected = np.zeros(np.shape(tp))
        for idx, t in enumerate(tp):
            n = np.linalg.norm(t)
            projected[idx, :] = t - np.dot(t, normal) * normal
            projected[idx, :] *= n / np.linalg.norm(projected[idx, :])

        return projected

    def _lat_lon_alts(self, loc):
        geo = sk.Geodetic(self._geoid_mode)

        lats = []
        lons = []
        alts = []
        for lo in loc:
            geo.from_xyz(lo)

            lats.append(geo.latitude)
            lons.append(geo.longitude)
            alts.append(geo.altitude)

        return np.asarray(lats), np.asarray(lons), np.asarray(alts)

    @property
    def altitudes(self):
        """
        Altitudes of the grid in m
        """
        return self._altitudes

    @property
    def numhoriz(self):
        """
        Number of horizontal divisions in the grid
        """
        return len(self._retprofileloc)

    def is_extended(self, angle_idx):
        """
        True if this angle index is outside the range of the input grid points.
        """
        if angle_idx < self._numextend:
            return True
        if angle_idx >= self.numhoriz - self._numextend:
            return True
        return False

    @property
    def numextended(self):
        """
        Number of extended grid points on each side of the grid
        """
        return self._numextend

    @property
    def grid_latitudes(self):
        """
        Latitudes of the grid
        """
        return self._lat_lon_alts(self._retprofileloc)[0]

    @property
    def grid_longitudes(self):
        """
        Longitudes of the grid
        """
        return self._lat_lon_alts(self._retprofileloc)[1]

    @property
    def image_latitudes(self):
        """
        Input image latitudes
        """
        return self._lat_lon_alts(self._averagescantp)[0]

    @property
    def image_longitudes(self):
        """
        Input image longitudes
        """
        return self._lat_lon_alts(self._averagescantp)[1]



from __future__ import annotations

from datetime import datetime

import numpy as np

import skretrieval.time

from .satellitebase import SatelliteBase
from .satellitekepler import SatelliteKepler
from .satellitesgp4 import SatelliteSGP4


# ------------------------------------------------------------------------------
#           class SatelliteKepler
# ------------------------------------------------------------------------------
class SatelliteSunSync(SatelliteBase):
    """
    Implements a sun-synchronous orbit. This model starts the chosen  orbit propagator with
    appropriate initial conditions given by the user. The orbit inclination is chosen by the code
    so the orbit with the given period (or altitude) is sun-synchronous. The user can utilize
    a Keplerian or SGP4 orbit propagator. The kepler propagator is useful for short, one orbit, studies but will not
    actually precess the orbit. The SGP4 propagator is slower but will capture the sun-synchronous precession over
    a period of days and months.
    """

    def __init__(
        self,
        utc: np.datetime64 | (datetime | (float | str)),
        orbittype="sgp4",
        sgp4_bstar_drag=0.0,
        **kwargs,
    ):
        """

        Parameters
        -----------
        utc: datetime.datetime
            The UTC time of the elements. This is set as the time of the ascending node of the selected orbit.

        period_from_seconds : float
            Optional, specifies the orbital period in seconds. An alternative to specify the period is with the optional
            parameter `period_from_altitude`. One, but only one, of the two optional methods must be used.

        period_from_altitude : float
             Optional, specifies the orbital period using the altitude of the satellite in meters. The altitude is nominal as we do not
             account for oblateness of the Earth etc. The radius of the Earth is internally assumed to be 6378000.0 meters.
             An alternative to specify the period is with the optional parameter `period_from_seconds`.  One, but only one, of the
             two optional methods must be used.

        orbittype: str
            (default='sgp4'). User can choose the orbit propagator. Currently accepted values are 'kepler' and 'sgp4'.

        sgp4_bstar_drag : float
            (default 0.0).  Allows the user to add a drag term. This value is only used by the SGP4 predictor and must
            be compatible with the SGP4 predictor.

        localtime_of_ascending_node_hours: float
            Specifies the nominal local time of the ascending node in hours (0-24).  If set then its value
            represents the "hour" at which you want the ascending node for example a floating point value of 18.25 will
            set the node to 18:15 LT. The current implementation may not be exact but will be close.

        longitude_of_ascending_node_degrees: float
            Specifies the geographic longitude of the ascending node (-180 to 360)at time `platform_utc`. This is an alternative method
            to using parameters `localtime_of_ascending_node_hours` and `right_ascension_ascending_node` to specify the right ascension of the ascending node.
            One, but only one, of the 3 optional methods may be used, if neither option is used the RAAN is set to 0.0.

        right_ascension_ascending_node: float
            Specifies the Right Ascension of the ascending node in radians (0 to :math:`2\\pi`). This is an alternative method
            to parameter `longitude_of_ascending_node_degrees` to specify the right ascension of the ascending node.
            One, but only one, of the two optional methods must be used, if neither option is used the RAAN is set to 0.0.

        argument_of_perigee: float
             Default 0. The argument of perigree in radians (0 to :math:`2\\pi`).

        eccentricity: float
            Default 0. The eccentricity of the orbit (0 to 1)

        orbitnumber: int
            Default 0. The orbit number at the epoch given by `platform_utc`


        """
        super().__init__()
        ut = skretrieval.time.ut_to_datetime(utc)
        self.satellite = None
        self.set_sun_sync_from_elements(
            ut, orbittype=orbittype, sgp4_bstar_drag=sgp4_bstar_drag, **kwargs
        )

    # -----------------------------------------------------------------------------
    #           update_eci_position
    # -----------------------------------------------------------------------------

    def update_eci_position(self, tnow: datetime):
        """
        Updates the eciposition and ecivelocity of the satellite at the given instant in time.  The user does not
        usually call this function directly. It is usually called by the base class :class:`~.SatelliteBase`

        Parameters
        ----------
        tnow : datetime
            Updates the ECI eci-position and ECI eci-velocity of the satellite to this time
        """
        self.satellite.update_eci_position(tnow)
        self._set_current_state(
            self.satellite.eciposition(), self.satellite.ecivelocity(), tnow
        )

    # ------------------------------------------------------------------------------
    #           period
    # ------------------------------------------------------------------------------
    def period(self) -> float:
        """
        Return the period of the satellite.  For low earth orbit spacecraft
        this is accurate to about 0.1 seconds.

        Returns
        -------
        datetime.timedelta
            The period of the satellite
        """
        return self.satellite.period()

    # -----------------------------------------------------------------------------
    #           eccentricity
    # -----------------------------------------------------------------------------

    def eccentricity(self) -> float:
        return self.satellite.eccentricity()

    # -----------------------------------------------------------------------------
    #           set_sun_sync_from_elements
    # -----------------------------------------------------------------------------

    def set_sun_sync_from_elements(
        self,
        utc: datetime,
        orbittype: str = "sgp4",
        sgp4_bstar_drag: float = 0.0,
        **kwargs,
    ):
        """
        Defines the sun-synchronous orbit using Keplerian orbital elements. The sun-synchronous orbit is always
        kick-started using a Kepler based orbit and it can continue to use that kepler orbit to propagate the position
        and velocity or it can use the SGP4 model. This function is called by the class instance constructor and the
        user does not normally need to call this function directly.

        Parameters
        ----------
        utc : datetime.datetime
            The universal time of the orbital elements.


        orbittype: str
            Specifies the type of orbit predictor to be used to propagate the orbit. It can be either 'kepler' or 'sgp4'. Default is 'sgp4'

        sgp4_bstar_drag: float
            Optional argument that is only used if the `sgp4` propagator is selected. This value is used as the bstar drag
            term in the SGP4 orbit predictor. The default value is 0.0 implying drag is ignored.

        kwargs: extra key word arguments.
            These keywords are from the keyword options in :meth:`.SatelliteKepler.from_elements`.
            Note that parameter `localtime_of_ascending_node` will override, if set, the `longitude_of_ascending_node_degrees`
            setting in the key words options.


        """

        kepler = SatelliteKepler(utc, inclination_is_sun_sync=True, **kwargs)
        if orbittype.lower() == "kepler":
            self.satellite = kepler
        elif orbittype.lower() == "sgp4":
            sgp4 = SatelliteSGP4()
            sgp4.from_kepler_orbit(kepler, bstar=sgp4_bstar_drag)
            self.satellite = sgp4
        else:
            msg = f"the requested type of orbit predictor {orbittype} is not supported"
            raise Exception(msg)



"""
The `mjd` module provides time conversion routines to and from Modified Julian Date. It provides convenient
functions which convert scalars, arrays or sequences of various time representations (strings, datetimes and numpy datetime64 into floats) into
corresponding scalars or arrays of modified julian date. The code uses the time functions native to python and makes
no attempt to implements IERS (Earth Rotation) details such as leap seconds etc. The following functions are provided,

 - `utc_to_mjd`. Converts scalars, arrays and sequences of various time representations to floating point MJD
 - `mjd_to_utc`, Converts scalars, arrays and sequences of floating point MJD to numpy.datetime64
 - `mjd_to_datetime64`, converts scalar mjd to numpy.datetime64
 - `mjd_to_datetime`, converts scalar mjd to datetime.datetime
 - `datetime64_to_datetime`, converts numpy.datetime64 to datetime.datetime
"""
from __future__ import annotations

import math
import numbers
import re
import string
import time
from collections.abc import Sequence
from datetime import datetime, timedelta

import numpy as np
from astropy.time import Time

_g_mjd0str = "1858-11-17 00:00:00.000000"
_t0_64 = np.datetime64(_g_mjd0str)
_t0 = datetime.strptime(_g_mjd0str, "%Y-%m-%d %H:%M:%S.%f")


# ------------------------------------------------------------------------------
#           datetime64_to_mjd
# ------------------------------------------------------------------------------
def datetime64_to_mjd(utc: np.datetime64) -> float:
    """
    Converts a numpy datetime64 to a floating point MJD. Only scalar values are supported
    by this function.

    Parameters
    ----------
    utc : numpy.datetime64
         A time specified with a numpy datetime64 object. Explicit time zones are
         not currently supported. Only single scalar values are supported in this function.

    Returns
    -------
    float
        The corresponding modified julian date.

    """
    return (
        (utc - _t0_64).astype("timedelta64[us]")
        / np.timedelta64(1, "us")
        / 86400000000.0
    )


# ------------------------------------------------------------------------------
#           mjd_to_datetime64
# ------------------------------------------------------------------------------
def mjd_to_datetime64(mjd: float) -> np.datetime64:
    """
    Converts an modified julian date to a numpy.datetime64. This function only works
    on scalar values of mjd.

    Parameters
    ----------
    mjd : float
        The modified julian date.

    Returns
    -------
    numpy.datetime64
        The numpy.datetime64 corresponding to the modified julian date
    """
    dt = np.timedelta64(int(mjd * 86400000000.0 + 0.5), "us")
    return (_t0_64 + dt).astype("datetime64[us]")


# ------------------------------------------------------------------------------
#           mjd_to_datetime64
# ------------------------------------------------------------------------------
def mjd_to_datetime(mjd: float) -> datetime:
    """
    Converts a modified julian date to a datetime.datetime object. This function only works
    on scalar values of mjd. Time zones are not set.

    Parameters
    ----------
    mjd : float
        The modified julian date.

    Returns
    -------
    datetime.datetime
        The datetime.datetime corresponding to the modified julian date. No time zone is set
    """

    dt = timedelta(days=mjd)
    return _t0 + dt


# ------------------------------------------------------------------------------
#           datetime_to_mjd
# ------------------------------------------------------------------------------
def datetime_to_mjd(t: datetime) -> float:
    """
    Converts a python datetime object to a floating point MJD. Only scalar values are supported
    by this function.

    Parameters
    ----------
    utc : datetime.datetime
         A time specified with a regular python datetime.datetime object. Explicit time zones are
         not currently supported. Only single scalar values are supported in this function.

    Returns
    -------
    float
        The corresponding modified julian date.

    """
    dt = t - _t0  # from datetime
    return (
        float(dt.days)
        + (float(dt.seconds) + float(dt.microseconds) * 1.0e-06) / 86400.0
    )  # to mjd as a floating point


# ------------------------------------------------------------------------------
#           mjd_to_utc
# ------------------------------------------------------------------------------
def mjd_to_ut(mjd):
    """
    A convenience function that converts scalars or arrays of floating point modified julian dates to universal time representations.
    The universal time is always returned as a scalar or array of numpy.datetime64 object(s).

    Parameters
    ----------
    mjd : scalar, array, or sequence of float
        The input can be a scalar, numpy array or regular python list/tuple of floats storing MJD values.

    Returns
    -------
    numpy.datetime64 or numpy.ndarray< numpy.datetime64 >
        Returns either a scalar or array of numpy.datetime64 that matches the input object
    """

    if (type(mjd) is np.ndarray) or (isinstance(mjd, Sequence)):
        if type(mjd) is np.ndarray:
            utc = np.zeros_like(mjd, dtype="datetime64[us]")
            with np.nditer(
                [mjd, utc],
                flags=["buffered"],  # create the nupy iterator object
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # and convert each elemnet
                    b[...] = mjd_to_datetime64(a)  # from datetime to mjd
                utc = it.operands[1]
        else:
            utc = np.zeros([len(mjd)])
            for i in range(len(mjd)):
                utc[i] = mjd_to_datetime64(mjd[i])
    else:
        utc = mjd_to_datetime64(mjd)
    return utc


# ------------------------------------------------------------------------------
#           datetime64_to_mjd
# ------------------------------------------------------------------------------
def ut_to_mjd(utc):
    """
    A convenience function that converts various arrays or scalar representations of UT to modified julian date.
    Scalar input values will return as a scalar float modified julian date while array, list or tuple input values are all returned
    as numpy arrays of mjd in float64.

    Parameters
    ----------

    utc : scalar, list, tuple or array
        The input time which represents a coordinated universal time. It can be represented by
        (i) a string in a supported python datetime.isoformat
        (ii) a number. The number is assumed to represent a modified julian date.
        (iii) a datetime.datetime object.
        (iv) a numpy.datetime64 object.
        The *utc* object can be a scalar, list, tuple or numpy array. The same representation format must be used for
        all objects in the arrays and sequences.

    Returns
    -------
    scalar or numpy.ndarray of float
        The mjd is returned as a scalar if the input was a scalar or as a numpy array for input sequences and arrays
    """

    if type(utc) is np.ndarray:
        # Handle an numpy array of MJD values, datetime64, datetime objects or mjd numbers
        if utc.dtype.kind == "M":  # if we have a numpy array of datetime64 object
            mjd = datetime64_to_mjd(utc)  # then convert it to floating point MJD
        elif (
            (utc.dtype.kind == "O") and (len(utc) > 0) and (type(utc[0]) is datetime)
        ):  # otherwiseif we have a numpy array of  datetime objects
            mjd = np.zeros_like(utc, dtype=np.float64)  # create the output array
            with np.nditer(
                [utc, mjd],
                flags=["refs_ok", "buffered"],  # create the nupy iterator object
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # and convert each elemnet
                    t = datetime_to_mjd(a)  # from datetime to mjd
                    b[...] = t
                mjd = it.operands[1]
        elif utc.dtype.kind == "U":  # otherwise if the array is strings
            mjd = np.zeros_like(
                utc, dtype=np.float64
            )  # then make an array to hold the answer
            with np.nditer(
                [utc, mjd],
                flags=["refs_ok", "buffered"],
                op_flags=[["readonly"], ["writeonly"]],
            ) as it:
                for a, b in it:  # For each element in the array
                    t = datetime.fromisoformat(str(a))  # convert the string to datetime
                    b[...] = datetime_to_mjd(t)  # and from datetime to mjd
                mjd = it.operands[1]
        else:  # otherwise we have an array of something else, I assume its array of numbers
            mjd = np.array(
                utc, dtype=np.float64
            )  # so try and convert the array of numbers to floating point MJD.

    elif isinstance(utc, str):  # if we have a single string
        t = datetime.fromisoformat(utc)  # then convert the string to datetime
        mjd = datetime_to_mjd(t)  # and from datetime to mjd

    elif isinstance(utc, Sequence):  # if we have a list or tuple but not numpy array
        newutc = np.array(utc)  # then convert the array to numpy array
        mjd = ut_to_mjd(newutc)  # and convert the numpy array to mjd
    else:  # Its not a numpy array, its not a sequence, so assume its a scalar
        if type(utc) is datetime:  # if its a datetime object
            mjd = datetime_to_mjd(utc)  # to mjd as a floating point
        elif type(utc) is np.datetime64:  # if we have a single numpy.datetime64
            mjd = float(datetime64_to_mjd(utc))  # then convert it
        elif isinstance(utc, numbers.Number):  # if we have a number
            mjd = float(utc)  # then assume its already an MJD
        else:  # otherwise
            mjd = math.nan
            msg = "utc_to_mjd, unsupported data type {}. Cannot convert to mjd".format(
                str(type(utc))
            )
            raise Exception(msg)
    return mjd


# ------------------------------------------------------------------------------
#           utc_to_astropy
# ------------------------------------------------------------------------------
def utc_to_astropy(utc):
    """
    A convenience function that converts list,  arrays or scalar representations of UTC to an astropy Time object.
    The function is a shallow wrapper for strings, datetime and datetime64 arrays as numpy arrays of mjd in float64.

    Parameters
    ----------

    utc : scalar, list, tuple or array
        The input time which represents a coordinated universal time. It can be represented by
        (i) a string in a supported python datetime.isoformat
        (ii) a number. The number is assumed to represent a modified julian date.
        (iii) a datetime.datetime object.
        (iv) a numpy.datetime64 object.
        The *utc* object can be a scalar, list, tuple or numpy array. The same representation format must be used for
        all objects in the arrays and sequences.

    Returns
    -------
    scalar or numpy.ndarray of float
        The mjd is returned as a scalar if the input was a scalar or as a numpy array for input sequences and arrays
    """

    formattype = None
    if (type(utc) is np.ndarray) and utc.dtype.kind in ["i", "u", "f"]:
        formattype = "mjd"
    elif isinstance(utc, Sequence):
        if isinstance(utc[0], numbers.Number):
            formattype = "mjd"
    elif isinstance(utc, numbers.Number):
        formattype = "mjd"

    return (
        Time(utc, format=formattype, scale="utc")
        if formattype is not None
        else Time(utc, scale="utc")
    )


# ---------------------------------------------------------------------------
# This is an old class from the early OSIRIS days in 2001. Its needs to be
# updated before it is used
# ----------------------------------------------------------------------------
class MJD:
    def __init__(self, mjd=0.0):
        self.m_mjd = mjd

    def from_date(self, year, month, day):
        self.from_utc(year, month, day, 0, 0, 0)

    def from_utc(self, year, month, day, hour, minutes, seconds):
        tupl = [
            int(year),
            int(month),
            int(day),
            int(hour),
            int(minutes),
            int(seconds),
            0,
            0,
            -1,
        ]
        secs = time.mktime(tupl) - time.timezone
        self.FromUnixSeconds(secs)

    def from_str(self, utcstr):
        pattern = f"[{string.punctuation}{string.whitespace}]+"  # Get all of the puctuation and whitespace charcaters
        s = re.split(
            pattern, utcstr
        )  # split the string with white space and punctuation
        t = [0, 1, 1, 0, 0, 0]  # get default values for fields not suplied by user
        n = min(len(s), 6)  # make sure we dont use more than 6 fields
        for i in range(n):
            t[i] = int(s[i])  # now fill in the values supplied by user
        self.from_utc(t[0], t[1], t[2], t[3], t[4], t[5])  # convert those times to UTC

    def as_unix_seconds(self):
        secs = (self.m_mjd - 40587.0) * 86400.0
        if secs < 0:
            secs = 0
        return secs

    def AsTimeTuple(self):
        return time.gmtime(self.as_unix_seconds())

    def FromUnixSeconds(self, secs):
        self.m_mjd = secs / 86400.0 + 40587.0

    def SetTo(self, mjd):
        self.m_mjd = mjd

    def FromSystem(self):
        self.FromUnixSeconds(time.time())

    def MJD(self):
        return self.m_mjd

    def as_utc_str(self, doticks=0):
        secs = (
            self.as_unix_seconds() + 0.0005
        )  # Round the seconds to the nearest millisecond
        tupl = time.gmtime(secs)  # Now get the time using the system call
        utcstr = time.strftime("%Y-%m-%d %H:%M:%S", tupl)  # Encode as a string
        if doticks:  # add the milliseconds if required
            ticks = int((secs - int(secs)) * 1000.0)
            utcstr = "%s.%03d" % (utcstr, ticks)
        return utcstr

    def AsDateStr(self):
        secs = (
            self.as_unix_seconds() + 0.0005
        )  # Round the seconds to the nearest millisecond
        tupl = time.gmtime(secs)  # Now get the time using the system call
        return time.strftime("%Y-%m-%d", tupl)  # Encode as a string

    def AsLocalTimeStr(self, formatstr="%H:%M"):
        secs = self.as_unix_seconds()
        tupl = time.localtime(secs)
        return time.strftime(formatstr, tupl)

    def __add__(self, other):
        self.m_mjd = self.m_mjd + float(other)
        return self

    def __sub__(self, other):
        self.mjd = self.m_mjd - float(other)
        return self

    def __mul__(self, other):
        self.m_mjd = self.m_mjd * float(other)
        return self

    def __div__(self, other):
        self.m_mjd = self.m_mjd / float(other)
        return self

    def __neg__(self):
        self.m_mjd = -self.m_mjd
        return self

    def __pos__(self):
        return self

    def __abs__(self):
        self.m_mjd = abs(self.m_mjd)
        return self

    def __int__(self):
        return int(self.m_mjd)

    def __long__(self):
        return int(self.m_mjd)

    def __float__(self):
        return float(self.m_mjd)

    def __str__(self):
        return self.as_utc_str(1)



from __future__ import annotations

import unittest

import numpy as np
import sasktran as sk

from skretrieval.core.lineshape import Gaussian, Rectangle
from skretrieval.core.platform.satellite import fake_satellite_scan
from skretrieval.legacy.core.sensor.spectrograph import Spectrograph
from skretrieval.legacy.core.simulator import MeasurementSimulator
from skretrieval.legacy.retrieval.ozone import OzoneRetrieval
from skretrieval.retrieval.rodgers import Rodgers


class TestRetrieval(unittest.TestCase):
    def test_retrieval_basic(self):
        alts = np.arange(10, 60.5, 1)
        axis, sun = fake_satellite_scan(altitudes=alts)

        atmo = sk.Atmosphere()
        atmo["air"] = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmo["ozone"] = sk.Species(sk.O3DBM(), sk.Labow())

        sensor = Spectrograph(
            np.arange(590, 610, 1),
            pixel_shape=Gaussian(fwhm=1, mode="linear"),
            vert_fov=Gaussian(fwhm=0.0001),
            horiz_fov=Rectangle(width=1, mode="constant"),
        )

        options = {}
        options["numordersofscatter"] = 1

        simulator = MeasurementSimulator(sensor, axis, atmo, options)

        simulator.engine.geometry.sun = sun

        measurement_l1 = simulator.calculate_radiance()

        rodgers = Rodgers(max_iter=1)
        labow = sk.Labow()

        altitudes = np.linspace(500, 99500, 100)
        ozone = labow.get_parameter(
            "SKCLIMATOLOGY_O3_CM3",
            latitude=0,
            longitude=0,
            mjd=54372,
            altitudes=altitudes,
        )

        atmo_sim = sk.Atmosphere()
        atmo_sim["air"] = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmo_sim["ozone"] = sk.Species(
            sk.O3DBM(), sk.ClimatologyUserDefined(altitudes, {"ozone": ozone * 0.5})
        )

        atmo_sim.wf_species = "ozone"

        sensor = Spectrograph(
            np.arange(590, 610, 1),
            pixel_shape=Gaussian(fwhm=1, mode="linear"),
            vert_fov=Gaussian(fwhm=0.0001),
            horiz_fov=Rectangle(width=1, mode="constant"),
        )
        ozone_ret = OzoneRetrieval(atmo_sim["ozone"])

        options = {}
        options["numordersofscatter"] = 1
        options["calcwf"] = 2
        options["wfheights"] = ozone_ret._retrieval_altitudes
        options["wfwidths"] = np.ones_like(ozone_ret._retrieval_altitudes) * 1000

        forward_model = MeasurementSimulator(sensor, axis, atmo_sim, options)

        forward_model.engine.geometry.sun = sun

        rodgers.retrieve(measurement_l1, forward_model, ozone_ret)



from __future__ import annotations

import numpy as np
import sasktran as sk


def geodetic() -> sk.Geodetic:
    return sk.Geodetic("wgs84")


def target_lat_lon_alt(
    los_vector: np.array, obs_position: np.array
) -> tuple[float, float, float]:
    """
    Parameters
    ----------
    los_vector : np.array
        3 Element look vector in ECI
    obs_position : np.array
        3 element observer position in ECI

    Returns
    -------
    Tuple[float, float, float]
        Tangent latitude/longitude/altitude if limb looking. Ground latitude/longitude if nadir looking
    """

    los = sk.LineOfSight(mjd=0, observer=obs_position, look_vector=los_vector)

    location = los.tangent_location()

    if location is None:
        location = los.ground_intersection(0)

    return location.latitude, location.longitude, location.altitude



from __future__ import annotations

from copy import copy

import numpy as np
import sasktran2 as sk2
import xarray as xr
from scipy.linalg import block_diag, toeplitz

from skretrieval.retrieval.tikhonov import two_dim_vertical_first_deriv

from . import StateVectorElement


class StateVectorElementConstituent(
    StateVectorElement, sk2.constituent.base.Constituent
):
    def __init__(
        self,
        constituent: sk2.constituent.base.Constituent,
        constituent_name: str,
        property_names: list[str],
        min_value=None,
        max_value=None,
        prior_influence=0,
        first_order_tikh=None,
        log_space=False,
        enabled=True,
    ):
        """
        A state vector element that is a sasktran2.constituent

        Parameters
        ----------
        constituent : sk2.constituent.base.Constituent
            The sasktran2 constituent
        constituent_name : str
            A name for the constituent
        property_names : list[str]
            Property names of the constituent that will be retrieved
        min_value : dict, optional
            Minimum values for the property names as a dictionary, by default {}
        max_value : dict, optional
            maximumum values for the property names as a dictionary, by default {}
        prior_influence : int, optional
            Prior influence for the property names as a dictionary, by default 0
        first_order_tikh : dict, optional
            First order tikhonov factors for the property names as a dictionary, by default {}
        log_space : bool, optional
            If true then the state elements will be rescaled to logarithmic space, by default False
        """
        if first_order_tikh is None:
            first_order_tikh = {}
        if max_value is None:
            max_value = {}
        if min_value is None:
            min_value = {}
        self._log_space = log_space
        self._constituent = constituent
        self._property_names = property_names
        self._constituent_name = constituent_name
        self._min_value = min_value
        self._max_value = max_value

        self._prior = copy(self.state())
        self._prior_influence = prior_influence
        self._first_order_tikh = first_order_tikh

        self._prior_dict = {}
        for property_name in self._property_names:
            self._prior_dict[property_name] = copy(
                getattr(self._constituent, property_name)
            )
        super().__init__(enabled)

    def state(self) -> np.array:
        data = []

        for property_name in self._property_names:
            data.append(getattr(self._constituent, property_name))

        if self._log_space:
            return np.log(np.hstack(data))
        return np.hstack(data)

    def lower_bound(self) -> np.array:
        data = []
        for property_name in self._property_names:
            x = getattr(self._constituent, property_name)

            bound = self._min_value.get(property_name, -np.inf)

            data.append(np.ones(len(x)) * bound)

        if self._log_space:
            return np.log(np.hstack(data))
        return np.hstack(data)

    def upper_bound(self) -> np.array:
        data = []
        for property_name in self._property_names:
            x = getattr(self._constituent, property_name)

            bound = self._max_value.get(property_name, np.inf)

            data.append(np.ones(len(x)) * bound)

        if self._log_space:
            return np.log(np.hstack(data))
        return np.hstack(data)

    def inverse_apriori_covariance(self) -> np.ndarray:
        prior_mats = []

        for property_name in self._property_names:
            val = getattr(self._constituent, property_name)

            corel = np.zeros_like(val)
            corel[0] = 1
            # corel[1] = 0.7
            # corel[2] = 0.5
            # corel[3] = 0.3
            # corel[4] = 0.1

            if self._log_space:
                prior_mats.append(
                    np.linalg.inv(
                        toeplitz(corel) * self._prior_influence[property_name] ** 2
                    )
                )
            else:
                prior_mats.append(
                    np.linalg.inv(
                        toeplitz(corel)
                        * np.outer(
                            self._prior_dict[property_name],
                            self._prior_dict[property_name],
                        )
                        * self._prior_influence[property_name] ** 2
                    )
                )

            if property_name == "vmr" and self._constituent_name == "so2":
                prior_mats[-1][:20, :20] *= 1e10
                prior_mats[-1][-40:, -40:] *= 1e10

            if property_name in self._first_order_tikh:
                gamma = two_dim_vertical_first_deriv(1, len(val)) * (
                    1 / self._first_order_tikh[property_name]
                )

                if not self._log_space:
                    gamma /= self._prior_dict[property_name][np.newaxis, :]

                prior_mats[-1] += gamma.T @ gamma

        return block_diag(*prior_mats)

    def apriori_state(self) -> np.array:
        return self._prior

    def name(self) -> str:
        return self._constituent_name

    def propagate_wf(self, radiance: xr.Dataset) -> xr.Dataset:
        if "extinction_per_m" in self._property_names:
            radiance = radiance.rename(
                {
                    f"wf_{self._constituent_name}_extinction": f"wf_{self._constituent_name}_extinction_per_m"
                }
            )
        wfs = []
        for property_name in self._property_names:
            wfs.append(
                radiance[f"wf_{self._constituent_name}_{property_name}"].rename(
                    {
                        radiance[f"wf_{self._constituent_name}_{property_name}"].dims[
                            0
                        ]: "x"
                    }
                )
            )

            if self._log_space:
                x = getattr(self._constituent, property_name)
                wfs[-1].values *= x[:, np.newaxis, np.newaxis, np.newaxis]

        return xr.concat(wfs, dim="x")

    def update_state(self, x: np.array):
        start = 0
        for property_name in self._property_names:
            current = getattr(self._constituent, property_name)
            property_length = len(current)
            if self._log_space:
                sv = np.exp(x[start : start + property_length])
                sv[np.isnan(sv)] = self._max_value[property_name]
            else:
                sv = x[start : start + property_length]
            if property_name in self._min_value:
                sv[sv < self._min_value[property_name]] = self._min_value[property_name]
            if property_name in self._max_value:
                sv[sv > self._max_value[property_name]] = self._max_value[property_name]

            self._constituent.__setattr__(property_name, sv)

            start += property_length

    def modify_input_radiance(self, radiance: xr.Dataset):
        return radiance

    def add_to_atmosphere(self, atmo: sk2.Atmosphere):
        return self._constituent.add_to_atmosphere(atmo)

    def register_derivative(self, atmo: sk2.Atmosphere, name: str):
        return self._constituent.register_derivative(atmo, name)



from __future__ import annotations

from .satellite_simplegeostationary import SatelliteSimpleGeostationary
from .satellitebase import SatelliteBase
from .satellitekepler import SatelliteKepler
from .satellitemolniya import SatelliteMolniya
from .satellitesgp4 import SatelliteSGP4
from .satellitesunsync import SatelliteSunSync
